//-----------------------------------------------------------
// Assembler class specialized as a MSX BASIC assembly coder
// created by Amaury Carvalho, 2019
//-----------------------------------------------------------

#include "assembly.h"
#include "kernel.h"
#include "symbols.h"
#include "float.h"
#include "gfx.h"

//----------------------------------------------------------------------------------------------

Assembler::Assembler() {
    headerType = Assembler::type_binary;
    errorMessage = "Undefined error";
    errorFound = false;
    noCtrlBreak = false;
    mathBCD = false;
    fastGFX = false;
    showLineNumber = false;
    ifCount = 0;
    forCount = 0;
    lineCount = 0;
    vdpActive = false;
    tileBank = 7;  // 0, 1, 2
    tileFont = 0;  // 0=default, 1=compile zanac style, 2=konami gradius 1 style, 3=konami gradius 2 style
    datasetDecoder = 0;  // 0=raw, 1=rle, 2=pletter
    datasetName = "DEFAULT";
    datasetEnabled = false;
    hasTileFont = false;
    hasDecoder = false;
    hasRleDecoder = false;
    hasPletterDecoder = false;
    hasTileFont1 = false;
    hasTileFont2 = false;
    hasTileFont3 = false;
}

bool Assembler::build(Parser *parser, char *filename) {
    if(!create(filename)) {
        errorMessage = "Cannot create output file";
        errorFound = true;
        return false;
    }

    this->parser = parser;

    buildSymbols();

    writeHeader();
    writeMainCode();
    writeSupportCode();
    writeSymbols();
    writeFooter();

    close();
    return !errorFound;
}

//----------------------------------------------------------------------------------------------

bool Assembler::create(char *filename) {
    if ((file = fopen(filename, "w"))) {
        return true;
    }
    return false;
}

void Assembler::write(string line) {
    for(unsigned int i = 0; i < line.size(); i++)
        if(line[i]=='\n')
            lineCount ++;
    lineCount ++;
    fprintf(file, "%s\n", line.c_str());
}

void Assembler::writeNewLine() {
    fprintf(file, "\n");
}

void Assembler::writeSeparator() {
    write(";--------------------------------------------------------");
}

void Assembler::close() {
    if(file) {
        fclose(file);
    }
}

//----------------------------------------------------------------------------------------------

void Assembler::buildSymbols() {
    TagNode *tag;
    Lexeme *lexeme;
    unsigned int i, t;

    symbols.clear();

    addSymbol("LET");
    addSymbol("BOOLEAN.IF");

    // build symbols list

    t = parser->tags.size();

    for(i = 0; i < t; i++) {
        tag = parser->tags[i];
        if(tag)
            buildSymbolsFromTag(tag);
    }

    // verify existence of special symbols

    t = symbols.size();

    for(i = 0; i < t; i++) {
        lexeme = symbols[i];
        if(lexeme) {
            if(lexeme->type == Lexeme::type_keyword && lexeme->name == "SCREEN" )
                vdpActive = true;
        }
    }

}

void Assembler::buildSymbolsFromTag(TagNode *tag) {
    ActionNode *action;
    unsigned int i, t;

    t = tag->actions.size();

    for(i = 0; i < t; i++) {
        action = tag->actions[i];
        if(action)
            buildSymbolsFromAction(action);
    }
}

void Assembler::buildSymbolsFromAction(ActionNode *action) {
    Lexeme *lexeme;
    ActionNode *nextAction;
    unsigned int i, t;

    if(action->isRemark)
        return;

    addSymbol(action);

    t = action->lexemes.size();

    for( i = 0; i < t; i++) {
        lexeme = action->lexemes[i];
        if(lexeme)
            addSymbol(lexeme);
    }

    t = action->actions.size();

    for(i = 0; i < t; i++) {
        nextAction = action->actions[i];
        if(nextAction)
            buildSymbolsFromAction(nextAction);
    }
}

void Assembler::addSymbol(string name) {
    Lexeme *lexeme = new Lexeme();
    lexeme->name = name;
    lexeme->value = name;
    lexeme->type = Lexeme::type_keyword;
    lexeme->subtype = Lexeme::subtype_any;
    symbols.push_back(lexeme);
}

void Assembler::addSymbol(Lexeme *lexeme) {
    Lexeme *check;
    unsigned int i, t;
    bool found = false;
    char name[30];

    if(lexeme->subtype == Lexeme::subtype_internal)
        return;

    t = symbols.size();

    for(i = 0; i < t; i++) {
        check = symbols[i];
        if(check)
            if(lexeme->type == check->type && lexeme->value == check->value && lexeme->type != Lexeme::type_literal) {
                found = true;
                break;
            }
    }

    if(found) {
        lexeme->name = check->name;
    } else {
        if(lexeme->type == Lexeme::type_identifier) {
            sprintf(name, "IDF_%u", t);
        } else if(lexeme->type == Lexeme::type_literal) {
            sprintf(name, "LIT_%u", t);
        } else if(lexeme->type == Lexeme::type_keyword) {
            sprintf(name, "KEY_%s", lexeme->value.c_str());
        } else {
            sprintf(name, "UNK_%u", t);
        }
        lexeme->name = name;

        symbols.push_back(lexeme);
    }
}

void Assembler::addSymbol(ActionNode *action) {
    Lexeme *lexeme;
    unsigned int i, t;
    bool found = false;

    t = symbols.size();

    for(i = 0; i < t; i++) {
        lexeme = symbols[i];
        if(lexeme)
            if(lexeme->type == Lexeme::type_keyword && lexeme->value == action->name) {
                found = true;
                break;
            }
    }

    if(!found) {
        lexeme = new Lexeme();
        lexeme->type = Lexeme::type_keyword;
        lexeme->value = action->name;
        lexeme->name = action->name;
        symbols.push_back(lexeme);
    }
}

//----------------------------------------------------------------------------------------------

void Assembler::writeHeader() {

    write( ";---------------------------------------------------------------------------------------------------------" );
    write( "; Source code converted by MSXBAS2ASM - MSX BASIC TO Z80 ASSEMBLY CONVERTER" );
    write( "; MSXBAS2ASM developed by Amaury Carvalho, 2019, Brazil" );
    write( "; http://launchpad.net/msxbas2asm" );
    write( ";---------------------------------------------------------------------------------------------------------" );
    writeNewLine();

    writeSeparator();
    write("; MSX BIOS DATA/FUNCTION POINTERS");
    writeSeparator();

    writeBiosCalls();
    writeNewLine();

    writeSeparator();
    write("; MSX BASIC DATA/FUNCTION POINTERS");
    writeSeparator();

    writeBasicCalls();
    writeNewLine();

    writeSeparator();
    write("; MATH PACK ROUTINES");
    writeSeparator();

    writeMathCalls();
    writeNewLine();

    writeSeparator();
    write("; SUPPORT MACROS");
    writeSeparator();

    if(headerType == Assembler::type_rom) {
        write( "COMPILE_TO_ROM: EQU 1" );
    } else if(headerType == Assembler::type_dos ) {
        write( "COMPILE_TO_DOS: EQU 1" );
    } else {
        write( "COMPILE_TO_BIN: EQU 1" );
    }

    write(support_macros);

    write(program_header);

}

void Assembler::writeBiosCalls() {
    write( bios_functions );
    writeNewLine();
}

void Assembler::writeBasicCalls() {

    write( basic_functions );
    writeNewLine();

}

void Assembler::writeMathCalls() {
    if(!mathBCD) {
        write("FAST_MATH: EQU 1");
        return;
    } else
        write("BCD_MATH: EQU 1");
    write( math_pack_functions );
    writeNewLine();
}

void Assembler::writeMainCode() {
    unsigned int i, t;

    write(program_routines);
    writeNewLine();

    write(program_main_code);

    t = parser->tags.size();

    for(i = 0; i < t; i++) {
        currentTag = parser->tags[i];
        if(currentTag)
            writeCodeFromTag(currentTag);
    }

    write(program_end_code);

}

void Assembler::writeCodeFromTag(TagNode *tag) {
    ActionNode *action;
    unsigned int i, t;
    char name[100];

    writeNewLine();
    sprintf(name, "TAG_%s:", tag->name.c_str());

    write(name);

    if(showLineNumber) {
        sprintf(name, "            set.line.number %6s      ; current line number", tag->name.c_str());
        write(name);
    }

    if(headerType != Assembler::type_rom && !noCtrlBreak) {
        write("            verify.break                ; verify CTRL+BREAK");
    }

    if(parser->has_traps) {
        write("            check.traps                 ; check traps");
    }

    t = tag->actions.size();

    for(i = 0; i < t; i++) {
        action = tag->actions[i];
        if(action)
            writeCodeFromAction(action);
    }
}

void Assembler::writeCodeFromAction(ActionNode *action) {
    Lexeme *lexeme, *lexeme2;
    ActionNode *nextAction, *prevAction, *action2;
    int i, t, i2, t2;
    char line[255];

    if(action->name == "DATA") {

        actionData.actions.push_back(action);
        return;

    } else if(action->name == "LOAD_SET") {

        datasetEnabled = true;
        nextAction  = action->actions[0];
        sprintf(line, "            ld hl, DATA_SET_%-20s ; pointer", nextAction->lexemes[0]->value.c_str());
        write(line);
        sprintf(line, "            call LOAD_SET               ; action call");
        write(line);
        action->isRemark = true;
        return;

    } else if(action->name.substr(0,5) == "DATA_") {

        if(action->name == "DATA_TILE_FONT")  {
            action2 = action->actions[0];
            lexeme2 = action2->lexemes[0];
            if(lexeme2->value == "1") {
                hasTileFont1 = true;
                hasDecoder = true;
                hasPletterDecoder = true;
            } else if(lexeme2->value == "2") {
                hasTileFont2 = true;
                hasDecoder = true;
                hasPletterDecoder = true;
            } else if(lexeme2->value == "3") {
                hasTileFont3 = true;
                hasDecoder = true;
                hasPletterDecoder = true;
            }
            hasTileFont = true;
        } else if(action->name == "DATA_DECODER") {
            action2 = action->actions[0];
            lexeme2 = action2->lexemes[0];
            if(lexeme2->value == "RLE") {
                hasRleDecoder = true;
                hasDecoder = true;
            } else if(lexeme2->value == "PLETTER") {
                hasPletterDecoder = true;
                hasDecoder = true;
            }
        }

        datasetEnabled = true;
        action->isRemark = true;
        if( (i = actionDataSet.actions.size()) ) {
            prevAction = actionDataSet.actions[i-1];
            if(action->name == prevAction->name) {
                t = action->actions.size();
                for(i = 0; i < t; i++) {
                    prevAction->actions.push_back(action->actions[i]);
                }
                return;
            }
        }
        actionDataSet.actions.push_back(action);
        return;

    } else if(action->name == "_EXPR_") {
        t = action->actions.size();
        for(i = 0; i < t; i++) {
            nextAction = action->actions[i];
            if(nextAction)
                writeCodeFromAction(nextAction);
        }
        return;

    } else if(action->name == "BOOLEAN.IF") {

        ifCount ++;
        ifStack.push(ifCount);

        sprintf(line, "   IF_%-10i   :                    ; start of IF command", ifStack.top());
        write( line );

        t = action->actions.size();

        for(i = 0; i < t; i++) {

            nextAction = action->actions[i];
            if(nextAction) {

                if( nextAction->name == "_COND_") {

                    // condition logical expression
                    t2 = nextAction->actions.size();
                    for(i2 = 0; i2 < t2; i2++)
                        writeCodeFromAction(nextAction->actions[i2]);

                    write(        "              call BOOLEAN.IF           ; verify IF condition result, out in A");
                    write(        "              or a                      ; cp 0 = false");
                    sprintf(line, "              jp z, ELSE_%-10i     ; if false, jump to ELSE actions", ifStack.top());
                    write( line );

                } else if( nextAction->name == "_THEN_") {

                    sprintf(line, "   THEN_%-10i :                    ; THEN actions", ifStack.top());
                    write( line );

                    // then actions
                    t2 = nextAction->actions.size();
                    for(i2 = 0; i2 < t2; i2++)
                        writeCodeFromAction(nextAction->actions[i2]);

                    if(ifStack.empty()) {
                        errorMessage = "ENDIF without IF statement";
                        errorFound = true;
                        return;
                    }

                    sprintf(line, "              jp ENDIF_%-10i       ; jump to END of IF command", ifStack.top());
                    write( line );

                    sprintf(line, "   ELSE_%-10i :                    ; ELSE actions", ifStack.top());
                    write( line );

                } else if( nextAction->name == "_ELSE_") {

                    // else actions
                    t2 = nextAction->actions.size();
                    for(i2 = 0; i2 < t2; i2++)
                        writeCodeFromAction(nextAction->actions[i2]);

                }

            }

        }

        if(ifStack.empty()) {
            errorMessage = "ENDIF without IF statement";
            errorFound = true;
            return;
        }

        sprintf(line, "   ENDIF_%-10i:                    ; end of IF command", ifStack.top());
        write( line );

        ifStack.pop();

        return;

    } else if(action->name == "FOR") {

        forCount ++;
        forStack.push(forCount);

        sprintf(line, "   FOR_%-10i:                      ; start of FOR command", forStack.top());
        write( line );

        t = action->actions.size();

        for(i = 0; i < t; i++) {

            nextAction = action->actions[i];
            if(nextAction) {

                if( nextAction->name == "_INIT_") {

                    t2 = nextAction->actions.size();
                    for(i2 = 0; i2 < t2; i2++)
                        writeCodeFromAction(nextAction->actions[i2]);

                    sprintf(line, "            jp FOR.WHILE_%-10i     ; jump to test if end of FOR", forStack.top());
                    write( line );

                } else if( nextAction->name == "_STEP_") {

                    sprintf(line, "   FOR.STEP_%-10i:                 ; STEP action", forStack.top());
                    write( line );

                    t2 = nextAction->actions.size();
                    for(i2 = 0; i2 < t2; i2++)
                        writeCodeFromAction(nextAction->actions[i2]);

                    //sprintf(line, "            ;jp FOR.WHILE_%-10i     ; jump to test if end of FOR", forStack.top());
                    //write( line );

                } else if( nextAction->name == "_TO_") {

                    sprintf(line, "   FOR.WHILE_%-10i:                ; test if end of FOR", forStack.top());
                    write( line );

                    t2 = nextAction->actions.size();
                    for(i2 = 0; i2 < t2; i2++)
                        writeCodeFromAction(nextAction->actions[i2]);

                    write(        "            call BOOLEAN.IF             ; verify IF condition result, out in A");
                    write(        "            or a                        ; cp 0 = false");
                    sprintf(line, "            jp z, ENDFOR_%-10i     ; end the loop if while condition is false", forStack.top());
                    write( line );
                    //sprintf(line, "            jp FOR.BODY_%-10i      ; jump to test if end of FOR", forStack.top());
                    //write( line );

                }

            }

        }

        sprintf(line, "   FOR.BODY_%-10i:                 ; start of FOR user code", forStack.top());
        write( line );

        // start of body code

        return;

    } else if(action->name == "NEXT") {

        // end of body code

        if(forStack.empty()) {
            errorMessage = "NEXT without FOR statement";
            errorFound = true;
            return;
        }

        sprintf(line, "              jp FOR.STEP_%-10i    ; repeat actions", forStack.top());
        write( line );

        sprintf(line, "   ENDFOR_%-10i:                   ; END of FOR command", forStack.top());
        write( line );

        forStack.pop();

        return;

    }

    t = action->actions.size();
    if(t) {
        i = t;
        do {
            i--;
            nextAction = action->actions[i];
            if(nextAction)
                writeCodeFromAction(nextAction);
        } while(i);
    }

    t = action->lexemes.size();

    if(action->name == "_PARM_") {
        if(t) {
            i = t;
            do {
                i --;
                lexeme = action->lexemes[i];
                if(lexeme) {
                    sprintf(line, "            ld hl, %-20s ; parameter", lexeme->name.c_str() ); //, action->parmNo+1);
                    write( line );
                    write(        "            push.parm");
                }
            } while(i);
        }
    } else if(action->name == "GOTO") {
        if(t) {
            sprintf(line, "            jp TAG_%-20s ; go to", action->lexemes[0]->value.c_str());
            write( line );
        }
    } else if(action->name == "GOSUB") {
        if(t) {
            sprintf(line, "            call TAG_%-20s ; gosub", action->lexemes[0]->value.c_str());
            write( line );
        }
    } else if(action->name == "_TIME") {

    } else if(action->name == "RETURN") {
        //if(t) {
        write("            ret ; return/gosub");
        //}
    } else if(action->name == "END") {
        write( "            jp end_pgm");
        //write( "            __call_bios BIOS_DSPFNK   ; turn on function keys display");
        //write( "            ret                  ; end" );
    } else {
        sprintf(line, "            call %-20s   ; action call", action->name.c_str());
        write(line);

    }

}

void Assembler::writeSymbols() {
    Lexeme *lexeme;
    ActionNode *action, *action2;
    unsigned int i, t, k, w, ti, tw, tt, tk;
    char line[200];
    bool first;

    t = symbols.size();

    variableCount = 0;

    write(variables_workarea_begin);

    k = 0;

    for(i = 0; i < t; i++) {
        lexeme = symbols[i];
        if(lexeme) {
            if(lexeme->type == Lexeme::type_literal) {
                writeSymbolLiteral(lexeme);

            } else if(lexeme->type == Lexeme::type_identifier) {
                writeSymbolIdentifier(lexeme);

            } else if(lexeme->type == Lexeme::type_keyword) {
                k ++;
            }
        }
    }

    writeNewLine();
    sprintf(line, "AFTER_LAST_VARIABLE:   equ VAR_STACK.POINTER + %i", (variableCount * 11) );
    write(line);

    write(variables_workarea_end);

    writeSeparator();
    write("; DATA SIMBOLS");
    writeSeparator();
    writeNewLine();

    write("DATA_ITEMS:");

    ti = actionData.actions.size();
    tt = 0;

    for(i = 0; i < ti; i++) {

        action = actionData.actions[i];
        tw = action->actions.size();

        for(w = 0; w < tw; w++) {

            action2 = action->actions[w];
            lexeme = action2->lexemes[0];
            if(lexeme) {
                sprintf(line, "   dw %s", lexeme->name.c_str() );
                write(line);
                tt++;
            }

        }
    }

    sprintf(line, "DATA_ITEMS_COUNT:   equ %i", tt );
    write(line);
    writeNewLine();

    write("DATA_SET_ITEMS_START:");

    ti = actionDataSet.actions.size();
    tk = 0;
    first = true;

    for(i = 0; i < ti; i++) {

        action = actionDataSet.actions[i];

        if(action->name == "DATA_SET") {
            if(first)
                first = false;
            else
                write("   db 0xFF      ; end of data set");
            action2 = action->actions[0];
            lexeme = action2->lexemes[0];
            datasetName = lexeme->value;
            sprintf(line, "DATA_SET_%s:", datasetName.c_str() );
            write(line);
            tk++;
        } else if(action->name == "DATA_DECODER") {
            action2 = action->actions[0];
            lexeme = action2->lexemes[0];
            if(lexeme->value == "RAW")
                datasetDecoder = 0;
            else if(lexeme->value == "RLE") {
                datasetDecoder = 1;
            } else if(lexeme->value == "PLETTER") {
                datasetDecoder = 2;
            } else
                datasetDecoder = 0;
            sprintf(line, "   db 0x01, %i   ; data decoder", datasetDecoder);
            write(line);
        } else if(action->name == "DATA_TILE_BANK") {
            action2 = action->actions[0];
            lexeme = action2->lexemes[0];
            if(lexeme->value == "0")
                tileBank = 1;
            else if(lexeme->value == "1")
                tileBank = 2;
            else if(lexeme->value == "2")
                tileBank = 4;
            else if(lexeme->value == "3")  // 0+1
                tileBank = 3;
            else if(lexeme->value == "4")  // 1+2
                tileBank = 6;
            else if(lexeme->value == "5")  // 0+2
                tileBank = 5;
            else if(lexeme->value == "6" || lexeme->value == "ALL")  // 0+1+2
                tileBank = 7;
            else
                tileBank = 0;
            sprintf(line, "   db 0x02, %i   ; data tile bank", tileBank);
            write(line);
        } else if(action->name == "DATA_TILE_FONT") {
            action2 = action->actions[0];
            lexeme = action2->lexemes[0];
            if(lexeme->value == "0")
                tileFont = 0;
            else if(lexeme->value == "1") {
                tileFont = 1;
            } else if(lexeme->value == "2") {
                tileFont = 2;
            } else if(lexeme->value == "3") {
                tileFont = 3;
            } else
                tileFont = 0;
            sprintf(line, "   db 0x03, %i   ; data tile font", tileFont);
            write(line);
        } else if(action->name == "DATA_TILE_LOAD_SET") {
            action2 = action->actions[0];
            lexeme = action2->lexemes[0];
            sprintf(line, "   db 0x05, %s, 0   ; data tile load set", lexeme->value.c_str());
            write(line);
        } else if(action->name == "DATA_TILE_LOAD_COLOR") {
            action2 = action->actions[0];
            lexeme = action2->lexemes[0];
            sprintf(line, "   db 0x06, %s, 0   ; data tile load color", lexeme->value.c_str());
            write(line);
        } else if(action->name == "DATA_SPRITE_LOAD_SET") {
            action2 = action->actions[0];
            lexeme = action2->lexemes[0];
            sprintf(line, "   db 0x07, %s, 0   ; data sprite load set", lexeme->value.c_str());
            write(line);
        } else if(action->name == "DATA_SPRITE_LOAD_COLOR") {
            action2 = action->actions[0];
            lexeme = action2->lexemes[0];
            sprintf(line, "   db 0x08, %s, 0   ; data sprite load color", lexeme->value.c_str());
            write(line);
        } else if(action->name == "DATA_SPRITE_LOAD_MULTI") {
            action2 = action->actions[0];
            lexeme = action2->lexemes[0];
            sprintf(line, "   db 0x09, %s, 0   ; data sprite load color", lexeme->value.c_str());
            write(line);
        } else if(action->name == "DATA_SCREEN_LOAD_SET") {
            action2 = action->actions[0];
            lexeme = action2->lexemes[0];
            sprintf(line, "   db 0x0A, %s, 0   ; data screen load set", lexeme->value.c_str());
            write(line);
        } else if(action->name == "DATA_TILE_LIMIT_ON") {
            sprintf(line, "   db 0x0B, 0x01    ; data tile limit on");
            write(line);
        } else if(action->name == "DATA_TILE_LIMIT_OFF") {
            sprintf(line, "   db 0x0B, 0x00    ; data tile limit off");
            write(line);
        } else {

            tw = action->actions.size();
            tt = 0;

            if(action->name == "DATA_TILE_SET") {
                write("   db 0x10      ; data tile set");
                sprintf(line, "   dw %i        ; size", tw);
                write(line);
            } else if(action->name == "DATA_TILE_COLOR") {
                write("   db 0x11      ; data tile color");
                sprintf(line, "   dw %i        ; size", tw);
                write(line);
            } else if(action->name == "DATA_SPRITE_SET") {
                write("   db 0x12      ; data sprite set");
                sprintf(line, "   dw %i        ; size", tw);
                write(line);
            } else if(action->name == "DATA_SPRITE_COLOR") {
                write("   db 0x13      ; data sprite mono color");
                sprintf(line, "   dw %i        ; size", tw);
                write(line);
            } else if(action->name == "DATA_SPRITE_MULTI") {
                write("   db 0x14      ; data sprite multi color");
                sprintf(line, "   dw %i        ; size", tw);
                write(line);
            } else if(action->name == "DATA_SCREEN_SET") {
                write("   db 0x15      ; data screen set");
                sprintf(line, "   dw %i        ; size", tw);
                write(line);
            } else if(action->name == "DATA_TILE_WALL") {
                write("   db 0x16      ; data tile wall");
                sprintf(line, "   dw %i        ; size", tw);
                write(line);
            } else if(action->name == "DATA_TILE_HOTSPOT") {
                write("   db 0x17      ; data tile hotspot");
                sprintf(line, "   dw %i        ; size", tw);
                write(line);
            } else if(action->name == "DATA_MUSIC") {
                write("   db 0x18      ; data music");
                sprintf(line, "   dw %i        ; size", tw);
                write(line);
            } else {
                write("   db 0x19      ; data unknown");
                sprintf(line, "   dw %i        ; size", tw);
                write(line);
            }

            for(w = 0; w < tw; w++) {

                action2 = action->actions[w];
                lexeme = action2->lexemes[0];
                if(lexeme) {
                    if((tt % 16)) {
                        sprintf(line, "%s, %s", line, lexeme->value.c_str() );
                    } else {
                        if(tt)
                            write(line);
                        sprintf(line, "   db %s", lexeme->value.c_str() );
                    }
                    tt++;
                }

            }
            if(tt)
                write(line);

        }

    }

    if(!first)
        write("   db 0xFF      ; end of data set");

    sprintf(line, "DATA_SET_ITEMS_COUNT:   equ %i", tk );
    write(line);
    writeNewLine();

}

void Assembler::writeSymbolLiteral(Lexeme *lexeme) {
    char line[255];
    int words[4];
    float floatValue = 0;
    double doubleValue = 0;

    writeNewLine();

    // VALTYP: %=2 $=3 !=4 #=8

    if(lexeme->subtype == Lexeme::subtype_string) {
        // 3 = string  (byte 03h, byte name_char1, byte name_char2, byte length, word address
        write("; string literal");
        sprintf(line, "%s: db 3, 0, 0, %i", lexeme->name.c_str(), (int)strlen(lexeme->value.c_str()) - 2);
        write(line);
        sprintf(line, "            dw %s_DATA, 0, 0", lexeme->name.c_str());
        write(line);
        write(        "            db 0");
        sprintf(line, "%s_DATA: db %s, 0", lexeme->name.c_str(), lexeme->value.c_str());
        write(line);
    } else if(lexeme->subtype == Lexeme::subtype_numeric) {
        // 2 = integer (byte 02h, byte name_char1, byte name_char2, word data)
        write("; numerical literal");
        sprintf(line, "%s:   db 2, 0, 0", lexeme->name.c_str());
        write(line);
        sprintf(line, "      dw 0, %s, 0, 0", lexeme->value.c_str());
        write(line);
    } else if(lexeme->subtype == Lexeme::subtype_single_decimal) {
        // 4 = single precision (byte 04h, byte name_char1, byte name_char2, word[2] = byte exponent, byte[3] data)
        write("; single decimal literal");
        sprintf(line, "%s:   db 4, 0, 0", lexeme->name.c_str());
        write(line);

        floatValue = stof(lexeme->value);
        if(mathBCD)
            float2BCD(floatValue, (int *) &words);   // convert floatValue to IEEE 754-1985 BCD single precision format
        else
            float2FloatLib(floatValue, (int *) &words);   // convert floatValue to IEEE 754-1985 single precision format

        sprintf(line, "      dw %i, %i, 0, 0", words[0], words[1]);
        write(line);
        sprintf(line, ";%s_DATA:   db '%s',0", lexeme->name.c_str(), lexeme->value.c_str());
        write(line);
    } else if(lexeme->subtype == Lexeme::subtype_double_decimal) {
        // 8 = double precision (byte 08h, byte name_char1, byte name_char2, word[4] = byte exponent, byte[7] data)
        write("; double decimal literal");
        sprintf(line, "%s:   db 8, 0, 0", lexeme->name.c_str());
        write(line);

        doubleValue = stod(lexeme->value);
        if(mathBCD)
            double2BCD(doubleValue, (int *)&words);   // convert doubleValue to IEEE 754-1985 BCD double precision format
        else
            double2FloatLib(doubleValue, (int *)&words);   // convert doubleValue to IEEE 754-1985 double precision format

        sprintf(line, "      dw %i, %i, %i, %i", words[0], words[1], words[2], words[3]);
        write(line);
        sprintf(line, ";%s_DATA:   db '%s',0", lexeme->name.c_str(), lexeme->value.c_str());
        write(line);
    } else if(lexeme->subtype == Lexeme::subtype_basic_string) {
        // 3 = string  (byte 03h, byte name_char1, byte name_char2, byte length, word address
        write("; BASIC string literal");
        sprintf(line, "%s:   db 3, 0, 0, %i", lexeme->name.c_str(), (int)strlen(lexeme->value.c_str()));
        write(line);
        sprintf(line, "      dw %s_DATA, 0, 0", lexeme->name.c_str());
        write(line);
        write(        "      db 0");
        sprintf(line, "%s_DATA: db '%s',0", lexeme->name.c_str(), lexeme->value.c_str());
        write(line);
    } else {
        write("; unknown literal");
        sprintf(line, "%s:   db 0, 0, 0, '%s', 0", lexeme->name.c_str(), lexeme->value.c_str());
        write(line);
    }

    // array variable type
    // https://github.com/Konamiman/MSX2-Technical-Handbook/blob/master/md/Chapter2.md
    // https://raw.githubusercontent.com/Konamiman/MSX2-Technical-Handbook/master/pics/Figure%202.9.png
    // array type = 1 byte
    // variable name = 1 word
    // length (bytes) of variable data = 1 word
    // number of dimensions = 1 byte
    // last index of 2nd dimension + 1 = 1 word
    // last index of 1st dimension + 1 = 1 word
    // variable data = n bytes, same structure of simple variables data

}

void Assembler::writeSymbolIdentifier(Lexeme *lexeme) {
    char line[255];

    if(lexeme->subtype == Lexeme::subtype_internal)
        return;

    writeNewLine();
    sprintf(line, "; identifier %s", lexeme->value.c_str());
    write(line);

    sprintf(line, "%s:   equ VAR_STACK.POINTER + %i", lexeme->name.c_str(), (variableCount * 11) );
    write(line);

    variableCount ++;
}

void Assembler::writeSymbolKeyword(Lexeme *lexeme) {
    if(lexeme)
        writeSymbolKeyword(lexeme->name);
}

void Assembler::writeSymbolKeyword(string name) {
    int i=0;
    char line[255];
    bool found;

    if(name[0] == '_' || name == "END") {
        return;
    }

    //writeSeparator();
    write("; keyword");
    sprintf(line, "%s:", name.c_str());
    write(line);

    found = false;

    while( existSymbolKeyword(i) ) {
        if( isSymbolKeyword(name) ) {
            found = true;
            write( getSymbolKeyword(i, 1) );
            break;
        }
        i++;
    }
    if(!found) {
        sprintf(line, "; abstract virtual %s", name.c_str());
        write(line);
        //write(getSymbolKeyword(SYMBOL_NOP, 1));	// nop
    }

}

void Assembler::writeInitializeSymbols() {
    Lexeme *lexeme;
    unsigned int i, t;

    t = symbols.size();
    variableCount = 0;

    write(variables_initialize);

    for(i = 0; i < t; i++) {
        lexeme = symbols[i];
        if(lexeme) {
            if(lexeme->subtype == Lexeme::subtype_internal)
                continue;
            if(lexeme->type == Lexeme::type_literal) {
                writeInitializeSymbolLiteral(lexeme);

            } else if(lexeme->type == Lexeme::type_identifier) {
                writeInitializeSymbolIdentifier(lexeme);

            }
        }
    }

    write("              ret");

}

void Assembler::writeInitializeSymbolLiteral(Lexeme *lexeme) {
    //char line[255];

    if(lexeme->subtype == Lexeme::subtype_string) {
    } else if(lexeme->subtype == Lexeme::subtype_numeric) {
    } else if(lexeme->subtype == Lexeme::subtype_double_decimal) {
    } else if(lexeme->subtype == Lexeme::subtype_basic_string) {
    } else if(lexeme->subtype == Lexeme::subtype_single_decimal) {
    } else {
    }
}

void Assembler::writeInitializeSymbolIdentifier(Lexeme *lexeme) {
    char line[255];
    int i = lexeme->value.c_str()[0] - 'A', vartype = 0;

    if(i>= 0 && i < 26)
        vartype = parser->deftbl[ i ];

    sprintf(line, "              ld hl, %s", lexeme->name.c_str());
    write(line);

    if(vartype) {
        sprintf(line, "              ld d, %i       ; explicitly coded defined type", vartype);
        write(line);
    } else {
        if(lexeme->subtype == Lexeme::subtype_string) {
            write(    "              ld d, 3       ; string");
        } else if(lexeme->subtype == Lexeme::subtype_numeric) {
            write(    "              ld d, 2       ; integer");
        } else if(lexeme->subtype == Lexeme::subtype_double_decimal) {
            write(    "              ld d, 8       ; double");
        } else if(lexeme->subtype == Lexeme::subtype_basic_string) {
            write(    "              ld d, 3       ; string");
        } else if(lexeme->subtype == Lexeme::subtype_single_decimal) {
            write(    "              ld d, 4       ; single");
        } else {
            write(    "              ld d, 2       ; any = default integer");
        }
    }

    sprintf(line, "              ld c, %-3i     ; variable name 1 (variable number)", (int)variableCount);
    write(line);
    if(lexeme->subtype == Lexeme::subtype_any && vartype == 0)
        sprintf(line, "              ld b, %-3i     ; variable name 2 (type flag=any)", 0);
    else
        sprintf(line, "              ld b, %-3i     ; variable name 2 (type flag=fixed)", 255);
    write(line);
    write(        "              call INIT_VAR ; variable initialize");

    variableCount ++;

}

void Assembler::double2BCD(double value, int *words) {
    int exponent, digit, i, k, neg;
    double mantissa;
    unsigned char digits[14], *p;
    for(i = 0; i < 4;  i ++)
        words[i] = 0;
    for(i = 0; i < 14; i ++)
        digits[i] = 0;
    neg = 64;
    if(value < 0) {
        neg += 128; // binary 10000000
        value = -value;
    }
    exponent = (int)log10(value);
    exponent ++;
    mantissa = pow(10, exponent);
    mantissa = value / mantissa;
    i = 0;
    while(mantissa > 0 && i < 14) {
        mantissa *= 10;
        digit = (int) mantissa;
        digits[i] = (unsigned char) digit;
        mantissa -= digit;
        i ++;
    }
    p = (unsigned char *) &words[0];            // word 0
    p[0] = neg + exponent;
    p[1] = digits[0] * 16 + digits[1];

    for(i = 1, k = 2; i < 4; i++, k += 4) {
        p = (unsigned char *) &words[i];            // word i
        p[0] = digits[k] * 16 + digits[k + 1];
        p[1] = digits[k + 2] * 16 + digits[k + 3];
    }
}

void Assembler::float2BCD(float value, int *words) {
    double d_value = value;
    int d_words[4];
    double2BCD(d_value, (int *) &d_words);
    words[0] = d_words[0];
    words[1] = d_words[1];
}

void Assembler::double2FloatLib(double value, int *words) {
    float value2 = value;
    float2FloatLib(value2, words);
}

void Assembler::float2FloatLib(float value, int *words) {
    unsigned short *pi = (unsigned short *) &value;
    unsigned char *pc = (unsigned char *) &value;
    unsigned char sign = 0;

    if(value == 0.0) {
        words[0] = 0;
        words[1] = 0;
        words[2] = 0;
        words[3] = 0;
        return;
    }

    if(pc[3] & 0x80)
        sign = 0x80;
    pc[3] <<= 1;

    if(pc[2] & 0x80)
        pc[3] |= 1;

    pc[2] &= 0x7F;
    pc[2] |= sign;

    pc[3] ++;

    words[0] = pi[0];
    words[1] = pi[1];
    words[2] = 0;
    words[3] = 0;

}

double Assembler::log10(double x) {
    return log(x)/log(10.0);
}

void Assembler::writeSupportCode() {

    writeBasicSupportCode();
    writeNewLine();

    write(variables_routines);
    writeNewLine();

    write(math16bits_routines);
    writeNewLine();

    write(memory_allocation_routines);
    writeNewLine();

    if(vdpActive) {
        if(fastGFX) {
            write(";FAST GRAPHICAL ROUTINES ACTIVATED (MSX1 SCREEN MODE ONLY)");
            write("GFX_FAST: equ 0x01");
            writeNewLine();
        }
        write(gfx_code);
        writeNewLine();

        if(hasTileFont) {
            write(font_data_routines);
            writeNewLine();
        }
    }

    if(hasDecoder) {
        write(compression_routines);
        writeNewLine();
    }

    if(!mathBCD) {
        write(float_wrapper);
        writeNewLine();
        write(float_code);
        writeNewLine();
    }

    writeInitializeSymbols();
    writeNewLine();

}

void Assembler::writeBasicSupportCode() {
    Lexeme *lexeme;
    unsigned int i, t;

    write(support_basic_begin);

    if(datasetEnabled) {
        write("EXIST_DATA_SET:   equ 1         ; data set to LOAD SET instruction included");

        if(hasRleDecoder)
            write("RLE_DECODER: equ 1");
        if(hasPletterDecoder)
            write("PLETTER_DECODER: equ 1");
        if(hasTileFont) {
            write("TILE_FONT_SUPPORT: equ 1");
            if(hasTileFont1)
                write("FONT_STYLE_1: equ 1");
            if(hasTileFont2)
                write("FONT_STYLE_2: equ 1");
            if(hasTileFont3)
                write("FONT_STYLE_3: equ 1");
        }

        writeNewLine();

        addSymbol("LOAD_SET");
    }

    t = symbols.size();

    for(i = 0; i < t; i++) {
        lexeme = symbols[i];
        if(lexeme) {
            if(lexeme->type == Lexeme::type_keyword) {
                writeSymbolKeyword(lexeme);

            }
        }
    }
    writeNewLine();

    write(support_basic_end);

}

void Assembler::writeFooter() {

    write(program_footer);

}

void Assembler::print() {
    Lexeme *lexeme;
    unsigned int i, t;

    t = symbols.size();

    for(i = 0; i < t; i++) {
        lexeme = symbols[i];
        if(lexeme) {
            if(lexeme->type == Lexeme::type_keyword)
                lexeme->indent = 0;
            else
                lexeme->indent = 2;
            lexeme->print();
        }
    }
}

void Assembler::error() {
    printf("%s\n", errorMessage.c_str());
}
