//-----------------------------------------------------------
// Lexer class specialized as a MSX BASIC tokenizer
// created by Amaury Carvalho, 2019
// reference: https://en.wikipedia.org/wiki/Lexical_analysis
//-----------------------------------------------------------

#include <stdio.h>
#include <string.h>
#include "lex.h"

Lexeme::Lexeme() {
    value = "";
    indent = 0;
    tag = 0;
    type = Lexeme::type_unknown;
    subtype = Lexeme::subtype_any;
}

void Lexeme::print() {
    printf("%*s--> %s %s: %s\n", indent, "", getTypeName(), getSubTypeName(), value.c_str());
}

char * Lexeme::getTypeName() {
    static char LexemeTypeName[][20] = { "Unknown", "Identifier", "Keyword", "Separator", "Operator", "Literal", "Comment" };

    return LexemeTypeName[type];
}

char * Lexeme::getSubTypeName() {
    static char LexemeSubTypeName[][20] = { "", "String", "Numeric", "Single Decimal", "Double Decimal", "Basic String", "Boolean/Remainder", "Function", "Internal" };

    return LexemeSubTypeName[subtype];
}

bool Lexeme::isKeyword() {
    // https://www.msx.org/wiki/Category:MSX-BASIC_Instructions
    return ( value == "CLS" || value == "PRINT" || value == "END" || value == "GOTO"
             || value == "IF" || value == "THEN" || value == "ELSE" || value == "COLOR"
             || value == "GOSUB" || value == "FOR" || value == "NEXT" || value == "RETURN"
             || value == "ON" || value == "OFF" || value == "STOP"
			 || value == "MOD" || value == "CALL" || value == "PAUSE"
			 || value == "INTERVAL" || value == "ERROR" || value == "FONT"
			 || value == "TILE" || value == "BANK" || value == "DECODER" || value == "INCLUDE" || value == "MULTI"
             || value == "COLLISION" || value == "COLLIDED" || value == "COLLIDER"
             || value == "MUSIC" || value == "WALL" || value == "HOTSPOT"
			 //|| value == "RAW" || value == "RLE" || value == "PLETTER"
             || value == "GET" || value == "DATE" || value == "TIME" || value == "SET"
             || value == "ASC" || value == "BIN" || value == "CDBL" || value == "CHR"
             || value == "CINT" || value == "CSNG" || value == "HEX" || value == "OCT"
             || value == "VAL" || value == "WIDTH" || value == "SPC" || value == "STR"
             || value == "BLOAD" || value == "BSAVE" || value == "CLOAD" || value == "CSAVE"
             || value == "CLOSE" || value == "EOF" || value == "LOAD" || value == "MAXFILES"
             || value == "MERGE" || value == "MOTOR" || value == "OPEN" || value == "RUN"
             || value == "SAVE" || value == "VARPTR" || value == "NEW"
             || value == "BASE" || value == "CIRCLE" || value == "COPY" || value == "SCREEN"
             || value == "CSRLIN" || value == "DRAW" || value == "LINE" || value == "LOCATE"
             || value == "LPOS" || value == "PAINT" || value == "POINT" || value == "POS"
             || value == "PRESET" || value == "PSET" || value == "DRAW" || value == "PEEK"
             || value == "POKE" || value == "VPEEK" || value == "VPOKE" || value == "INP"
             || value == "BASE" || value == "VDP" || value == "DEFINT" || value == "DEFDBL"
             || value == "DEFSNG" || value == "DEFSTR" || value == "DEF"
             || value == "OUT" || value == "WAIT" || value == "INKEY" || value == "INPUT"
             || value == "INPUT" || value == "KEY" || value == "STRIG" || value == "STEP"
             || value == "PAD" || value == "PDL" || value == "STICK" || value == "AND"
             || value == "EQV" || value == "IMP" || value == "NOT" || value == "OR"
             || value == "XOR" || value == "ABS" || value == "CDBL" || value == "CINT"
             || value == "CSNG" || value == "EXP" || value == "FIX" || value == "INT"
             || value == "LOG" || value == "RND" || value == "SGN" || value == "SQR"
             || value == "DEF" || value == "USR" || value == "LLIST" || value == "LPRINT"
             || value == "BEEP" || value == "PLAY" || value == "SOUND" || value == "SPRITE"
             || value == "SPRITE" || value == "INSTR" || value == "LEFT" || value == "MID"
             || value == "RIGHT" || value == "SPACE" || value == "STRING" || value == "ATN"
             || value == "COS" || value == "SIN" || value == "TAN" || value == "CLEAR"
             || value == "DATA" || value == "DIM" || value == "ERASE" || value == "LET"
             || value == "REDIM"
             || value == "REM" || value == "READ" || value == "RESTORE" || value == "RUN"
             || value == "SWAP" || value == "SLEEP" || value == "SEED" || value == "RANDOMIZE"
             || value == "IRND" || value == "SHR" || value == "SHL" || value == "INKEY"
             || value == "ISQR" || value == "PAUSE" || value == "KANJI" || value == "PUT" );
}

bool Lexeme::isBooleanOperator() {
    // https://www.msx.org/wiki/Category:MSX-BASIC_Instructions
    return ( value == "AND" || value == "OR" || value == "XOR" || value == "MOD"
             || value == "IMP" || value == "EQV" || value == "NOT" || value == "SHR" || value == "SHL" );
}

bool Lexeme::isFunction() {
    // https://www.msx.org/wiki/Category:MSX-BASIC_Instructions
    return ( value == "DATE" || value == "TIME" || value == "ASC" || value == "BIN"
             || value == "CDBL" || value == "CHR" || value == "SPC"
             || value == "CINT" || value == "CSNG" || value == "HEX" || value == "OCT"
             || value == "VAL" || value == "EOF" || value == "VARPTR" || value == "STR"
             || value == "CSRLIN" || value == "LPOS" || value == "POINT" || value == "POS"
             || value == "PRESET" || value == "PSET" || value == "PEEK"
             || value == "VPEEK" || value == "INP" || value == "BASE" || value == "VARPTR"
             || value == "OUT" || value == "INKEY" || value == "INPUT" || value == "INKEY"
             || value == "STRIG" || value == "PAD" || value == "PDL" || value == "STICK"
             || value == "ABS" || value == "CDBL" || value == "VDP" || value == "SNG"
             || value == "EXP" || value == "FIX" || value == "INT" || value == "DBL"
             || value == "LOG" || value == "RND" || value == "SGN" || value == "SQR"
             || value == "SPRITE" || value == "INSTR" || value == "LEFT" || value == "MID"
             || value == "RIGHT" || value == "SPACE" || value == "STRING" || value == "ATN"
             || value == "COS" || value == "SIN" || value == "TAN"
             || value == "COLLISION" || value == "COLLIDED" || value == "COLLIDER"
             || value == "TILE"
             || value == "IRND" || value == "ISQR"                    );
}

LexerLine::LexerLine() {
    lexemeIndex = 0;
}

void LexerLine::clearLexemes() {
    lexemes.clear();
}

void LexerLine::addLexeme(Lexeme *lexeme) {
    lexemes.push_back(lexeme);
}

void LexerLine::setLexemeBOF() {
    lexemeIndex = -1;
}

Lexeme *LexerLine::getCurrentLexeme() {
    return getLexeme(lexemeIndex);
}

Lexeme *LexerLine::getFirstLexeme() {
    return getLexeme(0);
}

Lexeme *LexerLine::getNextLexeme() {
    return getLexeme(lexemeIndex + 1);
}

Lexeme *LexerLine::getPreviousLexeme() {
    return getLexeme(lexemeIndex - 1);
}

Lexeme *LexerLine::getLastLexeme() {
    return getLexeme(lexemes.size() - 1);
}

Lexeme* LexerLine::getLexeme(int i) {
    if(i >= 0 && i < (int)lexemes.size()) {
        lexemeIndex = i;
        return lexemes[i];
    } else
        return 0;
}

int LexerLine::getLexemeCount() {
    return lexemes.size();
}

void LexerLine::pushLexeme() {
    lexemeStack.push(lexemeIndex);
}

void LexerLine::popLexeme() {
    if(!lexemeStack.empty()) {
        lexemeIndex = lexemeStack.top();
        lexemeStack.pop();
    }
}

void LexerLine::popLexemeDiscarting() {
    lexemeStack.pop();
}

void LexerLine::print() {
    printf("%s", line.c_str());
    for(unsigned int i=0; i < lexemes.size(); i++)
        lexemes[i]->print();
}

bool LexerLine::evaluate() {
    int i, t;
    char c;
    Lexeme *lexeme, *last_lexeme;

    lexemes.clear();

    t = line.length() - 1;

    lexeme = new Lexeme();
    last_lexeme = lexeme;

    for(i = 0; i < t; i++) {

        c = line[i];

        switch(lexeme->type) {
            case Lexeme::type_unknown: {
                    if( c <= ' ' ) {
                        continue;
                    } else if( isDecimal(c) || c == '"' || c == '&' ) {
                        lexeme->type = Lexeme::type_literal;
                        lexeme->value += c;
                        if(c == '.')
                            lexeme->subtype = Lexeme::subtype_double_decimal;
                        else if (c == '"')
                            lexeme->subtype = Lexeme::subtype_string;
                        else
                            lexeme->subtype = Lexeme::subtype_numeric;
                        continue;
                    } else if( isOperator(c) ) {
                        lexeme->type = Lexeme::type_operator;
                        lexeme->value += c;
                        lexemes.push_back( lexeme );
                        last_lexeme = lexeme;
                        lexeme = new Lexeme();
                        continue;
                    } else if( isSeparator(c) ) {
                        lexeme->type = Lexeme::type_separator;
                        lexeme->value += c;
                        lexemes.push_back( lexeme );
                        last_lexeme = lexeme;
                        lexeme = new Lexeme();
                        continue;
                    } else if( isIdentifier(c, true) ) {
                        lexeme->type = Lexeme::type_identifier;
                        lexeme->subtype = Lexeme::subtype_any;
                        lexeme->value += toupper(c);
                        lexeme->name = lexeme->value;
                        continue;
                    } else if( c == '$' && last_lexeme->type == Lexeme::type_keyword && last_lexeme->subtype == Lexeme::subtype_function ) {
                        last_lexeme->tag = 1;
                    } else {
                        lexeme->type = Lexeme::type_unknown;
                        lexeme->value += c;
                        lexemes.push_back( lexeme );
                        return false;
                    }
                }
                break;

            case Lexeme::type_literal: {
                    if( lexeme->subtype == Lexeme::subtype_string ) {
                        lexeme->value += c;
                        if( c == '"' ) {
                            lexemes.push_back( lexeme );
                            last_lexeme = lexeme;
                            lexeme = new Lexeme();
                        }
                        continue;

                    } else {
                        if( isDecimal(c) ) {
                            lexeme->value += c;
                            if( lexeme->subtype == Lexeme::subtype_double_decimal && c == '.' ) {    //strchr(lexeme->value.c_str(), '.' ) ) {
                                lexeme->type = Lexeme::type_unknown;
                                lexeme->subtype = Lexeme::subtype_any;
                                lexemes.push_back( lexeme );
                                return false;
                            } else {
                                if( c == '.' )
                                    lexeme->subtype = Lexeme::subtype_double_decimal;
                            }
                        } else if( isHexDecimal(c) ) {
                            if(lexeme->value.size() == 1 && (c == 'b' || c == 'B'))
                               c = 'x';
                            lexeme->value += c;
                        } else {
                            lexemes.push_back( lexeme );
                            last_lexeme = lexeme;
                            lexeme = new Lexeme();
                            if( c > ' '  )
                                i--;
                        }
                        continue;
                    }

                }
                break;

            case Lexeme::type_identifier: {
                    if( c <= ' ' || ! isIdentifier(c, false) ) {
                        if(lexeme->isKeyword()) {
                            lexeme->type = Lexeme::type_keyword;
                            lexeme->subtype = Lexeme::subtype_any;
                            if(lexeme->isBooleanOperator()) {
                                lexeme->type = Lexeme::type_operator;
                                lexeme->subtype = Lexeme::subtype_boolean_operator;
                            } else if(lexeme->isFunction()) {
                                lexeme->subtype = Lexeme::subtype_function;
                            }
                        }
                        lexemes.push_back( lexeme );
                        last_lexeme = lexeme;
                        lexeme = new Lexeme();
                        if( c > ' ')
                            i --;
                    } else {
                        lexeme->value += toupper(c);
                        lexeme->name = lexeme->value;
                        if(lexeme->isKeyword()) {
							bool isAnotherKeyword = false;
							for(int ii = i+1; ii < t && ii < (i+20); ii++) {
								c = line[ii];
								lexeme->value += toupper(c);
								if(lexeme->isKeyword()) {
									isAnotherKeyword = true;
									break;
								}
							}
							lexeme->value = lexeme->name;
							if(isAnotherKeyword)
								continue;
                            lexeme->type = Lexeme::type_keyword;
                            lexeme->subtype = Lexeme::subtype_any;
                            if(lexeme->isBooleanOperator()) {
                                lexeme->type = Lexeme::type_operator;
                                lexeme->subtype = Lexeme::subtype_boolean_operator;
                            } else if(lexeme->isFunction()) {
                                lexeme->subtype = Lexeme::subtype_function;
                            }
                            lexemes.push_back( lexeme );
                            if(lexeme->value == "REM")
                                return true;
                            last_lexeme = lexeme;
                            lexeme = new Lexeme();
                        } else {
                            // VALTYP: %=2 $=3 !=4 #=8
                            if     ( c == '%' )
                                lexeme->subtype = Lexeme::subtype_numeric;
                            else if( c == '$' )
                                lexeme->subtype = Lexeme::subtype_string;
                            else if( c == '!' )
                                lexeme->subtype = Lexeme::subtype_single_decimal;
                            else if( c == '#' )
                                lexeme->subtype = Lexeme::subtype_double_decimal;
                        }
                    }
                    continue;
                }
                break;

            case Lexeme::type_keyword: {
                    if( c <= ' ' || ! isIdentifier(c, false) ) {
                        lexemes.push_back( lexeme );
                        last_lexeme = lexeme;
                        lexeme = new Lexeme();
                        if( c > ' ')
                            i --;
                    } else
                        lexeme->value += toupper(c);
                    continue;
                }
                break;

            case Lexeme::type_operator: {
                    if( c <= ' ' || ! isOperator(c) ) {
                        lexemes.push_back( lexeme );
                        last_lexeme = lexeme;
                        lexeme = new Lexeme();
                        if( c > ' ' )
                            i --;
                    } else
                        lexeme->value += c;
                    continue;
                }
                break;

            case Lexeme::type_separator: {
                    if( c <= ' ' || ! isSeparator(c) ) {
                        lexemes.push_back( lexeme );
                        last_lexeme = lexeme;
                        lexeme = new Lexeme();
                        if( c > ' ' )
                            i --;
                    } else
                        lexeme->value += c;
                    continue;
                }
                break;

            case Lexeme::type_comment: {
                    if( c <= ' ' || ! isComment(c) ) {
                        lexemes.push_back( lexeme );
                        last_lexeme = lexeme;
                        lexeme = new Lexeme();
                        if( c > ' ' )
                            i --;
                    } else
                        lexeme->value += c;
                    continue;
                }
                break;

        }


    }

    if( lexeme->type != Lexeme::type_unknown ) {
        if(lexeme->isKeyword()) {
            lexeme->type = Lexeme::type_keyword;
            lexeme->subtype = Lexeme::subtype_any;
            if(lexeme->isBooleanOperator()) {
                lexeme->type = Lexeme::type_operator;
                lexeme->subtype = Lexeme::subtype_boolean_operator;
            } else if(lexeme->isFunction()) {
                lexeme->subtype = Lexeme::subtype_function;
            }
        }
        lexemes.push_back( lexeme );
    } else
        free(lexeme);

    return true;
}

bool LexerLine::isNumeric(char c) {
    return (c >= '0' && c <= '9');
}

bool LexerLine::isDecimal(char c) {
    return isNumeric(c) || c == '.';
}

bool LexerLine::isHexDecimal(char c) {
    return isNumeric(c) || c == 'B' || c == 'b' || c == 'h' || c == 'H' || c == 'o' || c == 'O' ||
           (c >= 'a' && c <= 'f') || (c >= 'A' && c <= 'F');
}

bool LexerLine::isSeparator(char c) {
    return (c == ':' || c == '(' || c == ')' || c == '{' || c == '}' || c == ',' || c == ';');
}

bool LexerLine::isOperator(char c) {
    return (c == '+' || c == '-' || c == '*' || c == '/' || c == '=' || c == '<' || c == '>' || c == '^' || c == '\\');
}

bool LexerLine::isIdentifier(char c, bool start) {
    return ( (c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z') || c == '_' ||
             ( ( c == '%' || c == '$' || c == '!' || c =='#' || (c >= '0' && c <= '9') ) && !start) );
}

bool LexerLine::isComment(char c) {
    return false;
}

bool Lexer::load(char *filename) {
    FILE *file;
    char line[255];
    int len = 255;
    LexerLine *lexerLine;

    lines.clear();

    if ((file = fopen(filename, "r"))) {
        //while ((getline(&line, &len, file)) != -1) {
        while ( fgets(line, len, file) ) {
            lexerLine = new LexerLine();
            lexerLine->line = line;
            lines.push_back( lexerLine );
        }

        fclose(file);

        return true;
    }
    return false;
}

bool Lexer::evaluate() {
    LexerLine *lexerLine;
    for(unsigned int i=0; i < lines.size(); i++) {
        lexerLine = lines[i];
        if(lexerLine)
            if(!lexerLine->evaluate()) {
                lineNo = i + 1;
                return false;
            }
    }
    return true;
}

void Lexer::print() {
    LexerLine *lexerLine;
    for(unsigned int i=0; i < lines.size(); i++) {
        lexerLine = lines[i];
        if(lexerLine)
            lexerLine->print();
    }
}

void Lexer::error() {
    LexerLine *lexerLine = lines[lineNo - 1];
    if(lexerLine)
        lexerLine->print();
}
