//******************************************************************************
//  PROJECT.......: MSXBAS2ASM
//  COPYRIGHT.....: Copyright (C) 2019- Amaury S P Carvalho
//  VERSION.......: 0.1
//  DESCRIPTION...: MSX Basic to Assembly converter
//  HOME PAGE.....: http://launchpad.net/msxbas2asm
//  CONTACT EMAIL.: amauryspires@gmail.com
//  HISTORY.......:
//  PERSON          DATE         COMMENT
//  --------        ----------   -----------------------------------------------
//  Amaury          27/05/2019   Initial version created
//******************************************************************************

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stddef.h>
#include <iostream>
#include <ctype.h>
#include <malloc.h>
#include <math.h>

#include "main.h"
#include "lex.h"
#include "parse.h"
#include "assembly.h"

using namespace std;

int main(int argc, char *argv[]) {
    int i;
    char *s;
    Lexer lexer;
    Parser parser;
    Assembler assembler;

    // parsing parameters

    for(i = 1; i < argc; i++) {
        if( argv[i][0] == '/' || argv[i][0] == '-' ) {
            switch( argv[i][1] ) {
                case 'h':
                case 'H':
                case '?':
                    parmHelp = true;
                    break;
                case 'd':
                case 'D':
                    parmDebug = true;
                    break;
                case 'q':
                case 'Q':
                    parmQuiet = true;
                    break;
                case 'c':
                case 'C':
                    parmCompile = true;
                    break;
                case 't':
                case 'T': {
                        parmBin = false;
                        parmRom = false;
                        parmDOS = false;
                        switch(argv[i][2]) {
                            case 'r':
                            case 'R':
                                parmRom = true;
                                break;
                            case 'd':
                            case 'D':
                                parmDOS = true;
                                break;
                            default:
                                parmBin = true;
                                break;
                        }
                    }
                    break;
                case '-': {
                        if(argv[i][2] == 'n' && argv[i][3] == 'c' && argv[i][4] == 'b' ) {
                            parmNoCtrlBreak = true;
                        } else if (argv[i][2] == 'b' && argv[i][3] == 'c' && argv[i][4] == 'd' ) {
							parmBCD = true;
                        } else if (argv[i][2] == 'f' && argv[i][3] == 'l' && argv[i][4] == 't' ) {
							parmBCD = false;
                        } else if (argv[i][2] == 'f' && argv[i][3] == 'l' && argv[i][4] == 't' ) {
							parmBCD = false;
                        } else if (argv[i][2] == 'd' && argv[i][3] == 'o' && argv[i][4] == 'c' ) {
							parmDoc = true;
                        } else if (argv[i][2] == 'v' && argv[i][3] == 'e' && argv[i][4] == 'r' ) {
							parmVer = true;
                        } else if (argv[i][2] == 'f' && argv[i][3] == 'g' && argv[i][4] == 'x' ) {
							parmFGX = true;
                        } else if (argv[i][2] == 'l' && argv[i][3] == 'i' && argv[i][4] == 'n' ) {
							parmLineNumber = true;
						}
                } break;
                default:
                    parmError = true;
                    break;
            }

        } else {

            strcpy(inputFilename, argv[i]);

        }
    }

    if(! parmQuiet) {
        // splash screen

        printf("%s %s\n", info_splash, app_version);

        // help hint
        if (parmDoc) {
			printf("%s", info_documentation);
			return 0;
        } else if (parmVer) {
			printf("%s", info_history);
			return 0;
		} else if(argc == 1 || parmHelp) {
            printf("%s", info_usage);
            return 0;
        } else {
            if(!parmBCD) {
              printf("Z80 float point library adapted from Zeda A.K. Thomas\n");
              printf("https://github.com/Zeda/z80float\n\n");
            }
        }
    } else {
        if (argc == 1 || parmHelp || parmError)
            return 1;
    }

    if(parmError) {
        printf("ERROR: Invalid parameter(s)\n");
        return 1;
    }

    if(parmCompile) {
        if(!FileExists(pasmoFilename)) {
            printf("ERROR: %s %s\n", pasmoFilename, pasmoError);
            return 1;
        }
    }

    if( ! inputFilename[0] ) {
        printf("ERROR: Input file name parameter is missing!\n");
        return 1;
    }

    if(! FileExists(inputFilename)) {
        printf("ERROR: Input file not found!\n");
        return 1;
    }

    // set output file name from input name

    strcpy(outputFilename, inputFilename);
    strcpy(compileFilename, inputFilename);

    s = strrchr(outputFilename, '.');
    if(s) {
        strcpy(s, ".asm");
    } else {
        strcat(outputFilename, ".asm");
    }

    s = strrchr(compileFilename, '.');
    if(s) {
        if(parmRom) {
            strcpy(s, ".rom");
        } else if(parmDOS) {
            strcpy(s, ".com");
        } else {
            strcpy(s, ".bin");
        }
    } else {
        if(parmRom) {
            strcat(compileFilename, ".rom");
        } else if(parmDOS) {
            strcat(compileFilename, ".com");
        } else {
            strcat(compileFilename, ".bin");
        }
    }

    if(FileExists(outputFilename)) {
        remove(outputFilename);
    }

    if(FileExists(compileFilename) && parmCompile) {
        remove(compileFilename);
    }

    if(! parmQuiet)
        printf("Converting %s to %s ...\n", inputFilename, outputFilename);

    // LEXICAL ANALYSIS
    // lexing the input file, tokenizing it

    if(! parmQuiet)
        printf("(1) Doing lexical analysis...\n");

    if( ! lexer.load(inputFilename) ) {
        printf("ERROR: Cannot load input file for lexical analysis\n");
        return 1;
    }

    if( ! lexer.evaluate() ) {
        lexer.error();
        printf("ERROR: Lexical error at line %i\n", lexer.lineNo);
        return 1;
    }

    if( parmDebug ) {
        printf("Displaying lexical analysis:\n");
        lexer.print();
    }

    // SYNTACTIC ANALYSIS
    // parsing the lexing tokens, building the syntax tree

    if(! parmQuiet)
        printf("(2) Doing syntactic analysis...\n");

    if( ! parser.evaluate(&lexer) ) {
        if(parser.lineNo == 0)
            printf("ERROR: Cannot build the symbol tree\n");
        else {
            parser.error();
            printf("ERROR: Syntax error at line %i\n", parser.lineNo);
        }
        return 1;
    }

    if( parmDebug ) {
        printf("Displaying syntactic analysis:\n");
        parser.print();
    }

    // SEMANTIC ANALYSIS
    // create assembly output

    if(! parmQuiet)
        printf("(3) Doing semantic analysis...\n");

    if(parmBin)
        assembler.headerType = Assembler::type_binary;
    else if(parmRom)
        assembler.headerType = Assembler::type_rom;
    else
        assembler.headerType = Assembler::type_dos;

    assembler.noCtrlBreak = parmNoCtrlBreak;
	assembler.mathBCD = parmBCD;
	assembler.fastGFX = parmFGX;
	assembler.showLineNumber = parmLineNumber;

    if( ! assembler.build(&parser, outputFilename) ) {
        assembler.error();
        printf("ERROR: Semantic analysis error\n");
        return 1;
    }

    if( parmDebug ) {
        printf("Displaying semantic analysis (symbol list):\n");
        assembler.print();
    }

    // finish process

    if(! parmQuiet)
        printf("Convertion finished with success.\n");

    // compiling...
    if(parmCompile) {
        char cmd[255];
        if(!parmQuiet)
            printf("(4) Compiling with Pasmo...\n");
        if(parmQuiet)
            sprintf(cmd, "%s %s %s", pasmoFilename, outputFilename, compileFilename);
        else
            if(parmDebug)
                sprintf(cmd, "%s -d -v %s %s", pasmoFilename, outputFilename, compileFilename);
            else
                sprintf(cmd, "%s -v %s %s", pasmoFilename, outputFilename, compileFilename);
        if(!system(cmd))
            printf("Compile process finished.\nFile %s created with success.\n", compileFilename);
        else {
            printf("Compiling error!\n");
            return 1;
        }
    }

    return 0;
}


bool FileExists(char *filename) {
    FILE *file;
    if ((file = fopen(filename, "r"))) {
        fclose(file);
        return true;
    }
    return false;
}

