//-------------------------------------------------------------------
// Parser class specialized as a MSX BASIC syntax tree builder
// created by Amaury Carvalho, 2019
// reference: https://en.wikipedia.org/wiki/Recursive_descent_parser
//            https://www.strchr.com/expression_evaluator
//            https://en.wikipedia.org/wiki/Shunting-yard_algorithm
//-------------------------------------------------------------------

#include "parse.h"

bool Parser::evaluate(Lexer *lexer) {
    int i, t = lexer->lines.size();
    LexerLine *lexerLine;

    has_traps = false;
    for(i = 0; i < 26; i++)
        deftbl[i] = 0;

    tags.clear();

    for(i = 0, lineNo = 1; i < t; i++, lineNo++) {
        lexerLine = lexer->lines[i];
        if(lexerLine->getLexemeCount() > 0) {
            line_comment = false;
            if(! eval_line(lexerLine))
                return false;
        }
    }

    return true;
}

bool Parser::eval_line(LexerLine *lexerLine) {
    Lexeme *lexeme = lexerLine->getFirstLexeme();
    LexerLine phrase;

    if_count = 0;
    error_line = lexerLine;

    if(lexeme) {
        if(lexeme->type == Lexeme::type_literal && lexeme->subtype == Lexeme::subtype_numeric) {

            tag = new TagNode();        // register line number tag
            tag->name = lexeme->value;
            tags.push_back(tag);

            while( (lexeme = lexerLine->getNextLexeme()) ) {

                if(lexeme->type == Lexeme::type_separator && lexeme->value == ":") {
                    if(if_count==0)
                        actionRoot = 0;
                    if(phrase.getLexemeCount())
                        if(!eval_phrase(&phrase))
                            return false;
                    phrase.clearLexemes();
                    continue;
                } else if( (lexeme->type == Lexeme::type_keyword && lexeme->name == "IF")
                           || (lexeme->type == Lexeme::type_keyword && lexeme->name == "ELSE") ) {
                    if(if_count==0)
                        actionRoot = 0;
                    if(phrase.getLexemeCount())
                        if(!eval_phrase(&phrase))
                            return false;
                    phrase.clearLexemes();
                    phrase.addLexeme(lexeme);
                    continue;
                } else {
                    phrase.addLexeme(lexeme);
                }

            }

            if(phrase.getLexemeCount()) {
                if(if_count == 0)
                    actionRoot = 0;
                if(!eval_phrase(&phrase))
                    return false;
            }

        } else {
            error_message = "Line number is missing";
            return false;
        }

    }

    while(if_count) {
        if_count --;
        popActionRoot();
    }

    return true;
}

bool Parser::eval_phrase(LexerLine *phrase) {
    Lexeme *lexeme = phrase->getFirstLexeme();

    if(lexeme) {
        if(lexeme->type == Lexeme::type_keyword) {
            if(!eval_statement(phrase))
                return false;
        } else if(lexeme->type == Lexeme::type_identifier) {
            phrase->setLexemeBOF();
            if(!eval_assignment(phrase))
                return false;
        } else {
            error_message = "Invalid keyword/identifier";
            return false;
        }
    }

    return true;
}

bool Parser::eval_statement(LexerLine *statement) {
    Lexeme *lexeme;
    ActionNode *action;
    bool result = true;

    lexeme = statement->getFirstLexeme();

    if(lexeme) {

        if     (lexeme->value == "REM")
            return eval_cmd_rem(statement);
        else if(lexeme->value == "DEF")
            return eval_cmd_def(statement, 0);
        else if(lexeme->value == "DEFINT")
            return eval_cmd_def(statement, 2);
        else if(lexeme->value == "DEFSTR")
            return eval_cmd_def(statement, 3);
        else if(lexeme->value == "DEFSNG")
            return eval_cmd_def(statement, 4);
        else if(lexeme->value == "DEFDBL")
            return eval_cmd_def(statement, 8);
        else if(lexeme->value == "CLEAR")
            return eval_cmd_rem(statement);
        else if(lexeme->value == "KEY")
            return eval_cmd_rem(statement);
        else if(lexeme->value == "LET")
            return eval_cmd_let(statement);
        else if(lexeme->value == "DIM")
            return eval_cmd_dim(statement);
        else if(lexeme->value == "REDIM")
            return eval_cmd_redim(statement);
        else if(lexeme->value == "ERASE")
            return eval_cmd_erase(statement);
        else if(lexeme->value == "PRINT")
            return eval_cmd_print(statement);
        else if(lexeme->value == "INPUT")
            return eval_cmd_input(statement);
        else if(lexeme->value == "READ")
            return eval_cmd_read(statement);
        else if(lexeme->value == "SCREEN")
            return eval_cmd_screen(statement);
        else if(lexeme->value == "SPRITE")
            return eval_cmd_sprite(statement);
        else if(lexeme->value == "BASE")
            return eval_cmd_base(statement);
        else if(lexeme->value == "VDP")
            return eval_cmd_vdp(statement);
        else if(lexeme->value == "PUT")
            return eval_cmd_put(statement);
        else if(lexeme->value == "TIME")
            return eval_cmd_time(statement);
        else if(lexeme->value == "THEN")
            return eval_cmd_then(statement);
        else if(lexeme->value == "ELSE")
            return eval_cmd_else(statement);
        else if(lexeme->value == "ON")
            return eval_cmd_on(statement);
        else if(lexeme->value == "INTERVAL")
            return eval_cmd_interval(statement);
        else if(lexeme->value == "STOP")
            return eval_cmd_stop(statement);
        else if(lexeme->value == "KEY")
            return eval_cmd_key(statement);
        else if(lexeme->value == "STRIG")
            return eval_cmd_strig(statement);
        else if(lexeme->value == "PLAY")
            return eval_cmd_play(statement);

        action = new ActionNode();
        action->name = lexeme->value;
        pushActionRoot(action);

        if     (lexeme->value == "CLS")
            result = eval_cmd_cls(statement);
        else if(lexeme->value == "DATA")
            result = eval_cmd_data(statement);
        else if(lexeme->value == "LOAD")
            result = eval_cmd_load(statement);
        else if(lexeme->value == "RESTORE")
            result = eval_cmd_restore(statement);
        else if(lexeme->value == "END")
            result = eval_cmd_end(statement);
        else if(lexeme->value == "BEEP")
            result = eval_cmd_beep(statement);
        else if(lexeme->value == "MUSIC")
            result = eval_cmd_music(statement);
        else if(lexeme->value == "GOTO")
            result = eval_cmd_goto(statement);
        else if(lexeme->value == "GOSUB")
            result = eval_cmd_gosub(statement);
        else if(lexeme->value == "RETURN")
            result = eval_cmd_return(statement);
        else if(lexeme->value == "LOCATE")
            result = eval_cmd_locate(statement);
        else if(lexeme->value == "DRAW")
            result = eval_cmd_draw(statement);
        else if(lexeme->value == "SOUND")
            result = eval_cmd_sound(statement);
        else if(lexeme->value == "COLOR")
            result = eval_cmd_color(statement);
        else if(lexeme->value == "IF")
            return eval_cmd_if(statement);
        else if(lexeme->value == "FOR")
            result = eval_cmd_for(statement);
        else if(lexeme->value == "NEXT")
            result = eval_cmd_next(statement);
        else if(lexeme->value == "SLEEP")
            result = eval_cmd_sleep(statement);
        else if(lexeme->value == "RANDOMIZE")
            result = eval_cmd_randomize(statement);
        else if(lexeme->value == "SEED")
            result = eval_cmd_seed(statement);
        else if(lexeme->value == "PSET")
            result = eval_cmd_pset(statement);
        else if(lexeme->value == "LINE")
            result = eval_cmd_line(statement);
        else if(lexeme->value == "CIRCLE")
            result = eval_cmd_circle(statement);
        else if(lexeme->value == "PAINT")
            result = eval_cmd_paint(statement);
        else if(lexeme->value == "WIDTH")
            result = eval_cmd_generic(statement, 1);
        else if(lexeme->value == "POKE")
            result = eval_cmd_poke(statement);
        else if(lexeme->value == "VPOKE")
            result = eval_cmd_vpoke(statement);
        else if(lexeme->value == "PAUSE")
            result = eval_cmd_generic(statement, 0);
        else
            result = false;

        popActionRoot();
    }

    return result;
}

bool Parser::eval_array_identifier(LexerLine *statement) {
    Lexeme *next_lexeme;
    LexerLine parm;
    int state = 0, sepCount = 1;

    while( (next_lexeme = statement->getNextLexeme()) ) {

        if(state == 0) {
            if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == "(") {
                sepCount ++;
            } else if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == ")") {
                sepCount --;
            }

            if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == ")" && sepCount == 0) {

                parm.setLexemeBOF();
                if(!eval_expression(&parm)) {
                    return false;
                }
                parm.clearLexemes();

                state ++;

            } else if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == ",") {

                parm.setLexemeBOF();
                if(!eval_expression(&parm)) {
                    return false;
                }
                parm.clearLexemes();

            } else {
                parm.addLexeme(next_lexeme);
            }

        } else
            break;

    }

    return true;
}

bool Parser::eval_assignment(LexerLine *assignment) {
    Lexeme *lexeme = assignment->getNextLexeme(), *next_lexeme;
    LexerLine parm;
    ActionNode *action, *actionExpr;
    bool result;

    if(lexeme) {
        if(lexeme->type == Lexeme::type_identifier || lexeme->subtype == Lexeme::subtype_function) {

            if(lexeme->subtype == Lexeme::subtype_function)
                lexeme->name = "_" + lexeme->name + "_";  // TIME = (expression)

            next_lexeme = assignment->getNextLexeme();
            if(next_lexeme) {

                action = new ActionNode();
                action->name = "LET";
                actionExpr = new ActionNode();
                actionExpr->name = "_PARM_";
                actionExpr->parmNo = 0;
                actionExpr->lexemes.push_back(lexeme);
                action->actions.push_back(actionExpr);
                pushActionRoot(action);

                if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == "(") {
                    // array assignment
                    eval_array_identifier(assignment);
                    next_lexeme = assignment->getCurrentLexeme();
                }

                if(next_lexeme->type == Lexeme::type_operator && next_lexeme->value == "=") {

                    parm.clearLexemes();
                    while( (next_lexeme = assignment->getNextLexeme()) ) {
                        parm.addLexeme(next_lexeme);
                    }
                    parm.setLexemeBOF();

                    result = eval_expression(&parm);

                    popActionRoot();

                    return result;

                } else
                    return false;

            } else
                return false;

            return true;

        } else
            return false;
    }
    return false;
}

bool Parser::eval_expression(LexerLine *parm) {
    stack <Lexeme*> operatorStack;
    Lexeme *lexeme, *next_lexeme;
    LexerLine functionLexemes;
    ActionNode *action, *actionRootSaved;
    int thisPreced, stackPreced;
    int outputCount = 0, sepcount=0;
    bool ok, unary = false, function = false, isRootPushed = false;

    if(parm->getLexemeCount() > 1 || !actionRoot)  {
        action = new ActionNode();
        action->name = "_EXPR_";
        action->parmNo = parmNo;
        pushActionRoot(action);
        isRootPushed = true;
    }

    lexeme = parm->getNextLexeme();

    while(lexeme) {

        // if operators, push to operators stack

        if(lexeme->type == Lexeme::type_operator || lexeme->type == Lexeme::type_separator) {

            // test when first operand in expression or unary operator

            if( outputCount == 0 || unary ) {

                if(outputCount == 0 && lexeme->value == ")") {
                    eval_expr_error = true;
                    error_message = "Mismatched parentheses error";
                    return false;
                } else if(lexeme->value == "=") {
                    next_lexeme = operatorStack.top();
                    if(next_lexeme->value == "<") {
                        next_lexeme->value = "<=";
                        lexeme = parm->getNextLexeme();
                        continue;
                    } else if(next_lexeme->value == ">") {
                        next_lexeme->value = ">=";
                        lexeme = parm->getNextLexeme();
                        continue;
                    } else {
                        eval_expr_error = true;
                        error_message = "Invalid = symbol in expression";
                        return false;
                    }
                } else if(lexeme->value == ">") {
                    next_lexeme = operatorStack.top();
                    if(next_lexeme->value == "<") {
                        next_lexeme->value = "<>";
                        lexeme = parm->getNextLexeme();
                        continue;
                    } else {
                        eval_expr_error = true;
                        error_message = "Invalid > symbol in expression";
                        return false;
                    }
                } else if(lexeme->value != "+" && lexeme->value != "-"
                          && lexeme->value != "(" && lexeme->value != ")"
                          && lexeme->value != "NOT") {
                    eval_expr_error = true;
                    error_message = "Invalid expression unary symbol";
                    return false;
                } else {
                    if(lexeme->value == "+") {  // unary plus
                        lexeme = parm->getNextLexeme();
                        continue;
                    } else if(lexeme->value == "-") {
                        lexeme->value = "_";  // unary minus
                    }
                }

            }

            if(lexeme->value == "(") {
                if(function) {

                    next_lexeme = operatorStack.top();
                    actionRootSaved = actionRoot;
                    actionRoot = pushActionFromLexeme( next_lexeme );
                    operatorStack.pop();

                    // parse function parameters
                    functionLexemes.clearLexemes();
                    ok = false;
                    sepcount = 0;
                    while( (lexeme = parm->getNextLexeme()) ) {
                        if( lexeme->value == "(" ) {
                            sepcount ++;
                        } else if( lexeme->value == ")" ) {
                            if(sepcount)
                                sepcount --;
                            else {
                                ok = true;
                                functionLexemes.setLexemeBOF();
                                if( ! eval_expression(&functionLexemes) )
                                    return false;
                                functionLexemes.clearLexemes();
                                break;
                            }
                        } else if(lexeme->value == ",") {
                            functionLexemes.setLexemeBOF();
                            if( ! eval_expression(&functionLexemes) )
                                return false;
                            functionLexemes.clearLexemes();
                            continue;
                        }
                        functionLexemes.addLexeme(lexeme);
                    }
                    if( ! ok ) {
                        eval_expr_error = true;
                        error_message = "Mismatched parentheses error in function";
                        return false;
                    }

                    actionRoot = actionRootSaved;

                    function = false;

                } else {
                    operatorStack.push(lexeme);
                }

                lexeme = parm->getNextLexeme();
                unary = false;
                continue;
            }

            if(function) {
                next_lexeme = operatorStack.top();
                pushActionFromLexeme( next_lexeme );
                operatorStack.pop();
                function = false;
            }

            // test this operator precedence against operators in stack

            thisPreced = getOperatorPrecedence(lexeme);

            while( ! operatorStack.empty() ) {
                next_lexeme = operatorStack.top();
                stackPreced = getOperatorPrecedence(next_lexeme);
                if( ( thisPreced <= stackPreced || next_lexeme->type == Lexeme::type_keyword ) && next_lexeme->value != "(" ) {
                    pushActionFromLexeme( next_lexeme );
                    operatorStack.pop();
                } else
                    break;
            }

            unary = true;

            if( lexeme->value == ")" ) {
                unary = false;
                ok = false;
                while( ! operatorStack.empty() ) {
                    next_lexeme = operatorStack.top();
                    if( next_lexeme->value == "(" ) {
                        operatorStack.pop();
                        ok = true;
                        break;
                    } else {
                        pushActionFromLexeme( next_lexeme );
                        operatorStack.pop();
                    }
                }
                if(!ok) {
                    eval_expr_error = true;
                    error_message = "Mismatched parentheses error";
                    return false;       // parentheses is missing
                }
            } else
                operatorStack.push(lexeme);

        } else if(lexeme->type == Lexeme::type_keyword) {

            operatorStack.push(lexeme);
            outputCount ++;
            function = true;
            //unary = true;

        } else {
            // else, push operand to output stack

            if(function) {
                next_lexeme = operatorStack.top();
                pushActionFromLexeme( next_lexeme );
                operatorStack.pop();
                function = false;
            }

            pushActionFromLexeme(lexeme);
            outputCount ++;
            unary = false;
            function = false;

        }

        lexeme = parm->getNextLexeme();

    }

    if(function) {
        next_lexeme = operatorStack.top();
        pushActionFromLexeme( next_lexeme );
        operatorStack.pop();
        function = false;
    }

    // if operators stack is not empty, copy to output stack
    while( ! operatorStack.empty() ) {
        pushActionFromLexeme( operatorStack.top() );
        operatorStack.pop();
    }

    if(isRootPushed)
        popActionRoot();

    return true;
}

int Parser::getOperatorPrecedence(Lexeme *lexeme) {
    int result;


    if     (lexeme->value == "NOT")
        result = 1;
    else if(lexeme->value == "AND")
        result = 2;
    else if(lexeme->value == "XOR")
        result = 3;
    else if(lexeme->value == "EQV")
        result = 4;
    else if(lexeme->value == "IMP")
        result = 5;
    else if(lexeme->value == "OR")
        result = 6;
    else if(lexeme->value == ">" || lexeme->value == "<" || lexeme->value == "=" || lexeme->value == ">=" || lexeme->value == "<=" )
        result = 7;
    else if(lexeme->value == "+" || lexeme->value == "-")
        result = 8;
    else if(lexeme->value == "*" || lexeme->value == "/" || lexeme->value == "\\"
            || lexeme->value == "MOD" || lexeme->value == "SHR" || lexeme->value == "SHL")
        result = 9;
    else if(lexeme->value == "^")
        result = 10;
    else if(lexeme->value == "_")  // unary minus
        result = 11;
    else
        result = 0;

    return result;
}

ActionNode * Parser::pushActionFromLexeme(Lexeme *lexeme) {
    ActionNode *actionExpr = 0;

    if(lexeme->type == Lexeme::type_operator || lexeme->type == Lexeme::type_separator) {

        actionExpr = new ActionNode();

        if     (lexeme->value == "+")
            actionExpr->name = "MATH.ADD";
        else if(lexeme->value == "-")
            actionExpr->name = "MATH.SUB";
        else if(lexeme->value == "/")
            actionExpr->name = "MATH.DIV";
        else if(lexeme->value == "*")
            actionExpr->name = "MATH.MULT";
        else if(lexeme->value == "\\")
            actionExpr->name = "MATH.IDIV";
        else if(lexeme->value == "^")
            actionExpr->name = "MATH.POW";
        else if(lexeme->value == "MOD")
            actionExpr->name = "MATH.MOD";
        else if(lexeme->value == "_")
            actionExpr->name = "MATH.NEG";
        else if(lexeme->value == ">")
            actionExpr->name = "BOOLEAN.GT";
        else if(lexeme->value == "<")
            actionExpr->name = "BOOLEAN.LT";
        else if(lexeme->value == ">=")
            actionExpr->name = "BOOLEAN.GE";
        else if(lexeme->value == "<=")
            actionExpr->name = "BOOLEAN.LE";
        else if(lexeme->value == "<>")
            actionExpr->name = "BOOLEAN.NE";
        else if(lexeme->value == "=")
            actionExpr->name = "BOOLEAN.EQ";
        else if(lexeme->value == "NOT")
            actionExpr->name = "BOOLEAN.NOT";
        else if(lexeme->value == "AND")
            actionExpr->name = "BOOLEAN.AND";
        else if(lexeme->value == "XOR")
            actionExpr->name = "BOOLEAN.XOR";
        else if(lexeme->value == "OR")
            actionExpr->name = "BOOLEAN.OR";
        else if(lexeme->value == "IMP")
            actionExpr->name = "BOOLEAN.IMP";
        else if(lexeme->value == "EQV")
            actionExpr->name = "BOOLEAN.EQV";
        else if(lexeme->value == "SHR")
            actionExpr->name = "BOOLEAN.SHR";
        else if(lexeme->value == "SHL")
            actionExpr->name = "BOOLEAN.SHL";
        else
            actionExpr->name = "_NOP_";

        actionRoot->actions.push_back(actionExpr);

    } else if(lexeme->type == Lexeme::type_keyword) {     // treatment of functions

        actionExpr = new ActionNode();
        actionExpr->name = lexeme->name;
        if(lexeme->name == "INPUT") {
            if(lexeme->tag == 1)
                actionExpr->name = "INPUT.FUNCTION.STR";
            else
                actionExpr->name = "INPUT.FUNCTION.INT";
        } else if(lexeme->name == "INKEY") {
            if(lexeme->tag == 1)
                actionExpr->name = "INKEY.STR";
        }

        actionRoot->actions.push_back(actionExpr);

    } else {

        if(actionRoot->isDirectParm) {
            actionRoot->lexemes.push_back(lexeme);
        } else {
            actionExpr = new ActionNode();
            actionExpr->name = "_PARM_";
            actionExpr->parmNo = parmNo;
            actionExpr->lexemes.push_back(lexeme);
            actionRoot->actions.push_back(actionExpr);
        }

    }

    return actionExpr;
}

//-----------------------------------------------------------------------------------------------------------------

bool Parser::eval_cmd_generic(LexerLine *statement, int parmCount) {
    Lexeme *next_lexeme;
    LexerLine parm;
    int i;

    // get keyword parameters
    i = 0;

    while( (next_lexeme = statement->getNextLexeme()) ) {

        if(next_lexeme->type == Lexeme::type_separator && ( next_lexeme->value == "," || next_lexeme->value == ";")) {
            i++;
            parm.setLexemeBOF();
            if(!eval_expression(&parm)) {
                return false;
            }
            parm.clearLexemes();
        } else {
            parm.addLexeme(next_lexeme);
        }

    }

    if(parm.getLexemeCount()) {
        i++;
        parm.setLexemeBOF();
        if(!eval_expression(&parm)) {
            return false;
        }
    }

    // compare parameters against expected

    if(parmCount >= 0) {
        if(parmCount == 0 && i > 0) {
            error_message = "No parameters allowed in this keyword";
            return false;

        } else if( i != parmCount ) {

            error_message = "Invalid number of parameters";
            return false;

        }
    }

    return true;
}

bool Parser::eval_cmd_let(LexerLine *statement) {
    return eval_assignment(statement);
}

bool Parser::eval_cmd_rem(LexerLine *statement) {
    return true;
}

bool Parser::eval_cmd_cls(LexerLine *statement) {
    return true;
}

bool Parser::eval_cmd_dim(LexerLine *statement) {
    Lexeme *next_lexeme;
    LexerLine parm;
    ActionNode *action = 0, *actionParm;
    int state = 0, sepCount = 0, dimCount=0;

    // get keyword parameters
    while( (next_lexeme = statement->getNextLexeme()) ) {

        if(state == 0) {
            if(next_lexeme->type == Lexeme::type_identifier) {
                action = new ActionNode();
                action->name = "ERASE";
                actionParm = new ActionNode();
                actionParm->name = "_PARM_";
                actionParm->lexemes.push_back(next_lexeme);
                action->actions.push_back(actionParm);
                pushActionRoot(action);
                dimCount = 0;
                state ++;
                sepCount = 0;
            } else
                return false;
        } else if(state == 1) {
            if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == "(") {
                sepCount ++;
                state ++;
            } else if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == ",") {
                popActionRoot();
                action = 0;
                state = 0;
            } else
                return false;

        } else if(state == 2) {
            if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == "(") {
                sepCount ++;
            } else if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == ")") {
                sepCount --;
            }

            if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == ")" && sepCount == 0) {

                parm.setLexemeBOF();
                if(!eval_expression(&parm)) {
                    return false;
                }
                parm.clearLexemes();

                dimCount ++;

                if(dimCount == 1)
                    actionRoot->name = "DIM1";
                else if(dimCount == 2)
                    actionRoot->name = "DIM2";
                else
                    return false;

                popActionRoot();

                state ++;

            } else if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == ",") {

                dimCount ++;

                parm.setLexemeBOF();
                if(!eval_expression(&parm)) {
                    return false;
                }
                parm.clearLexemes();

            } else {
                parm.addLexeme(next_lexeme);
            }

        } else {
            if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == ",") {
                popActionRoot();
                action = 0;
                state = 0;
            } else
                return false;
        }

    }

    if(action)
        popActionRoot();

    return true;
}

bool Parser::eval_cmd_redim(LexerLine *statement) {
    Lexeme *next_lexeme;
    LexerLine parm;
    ActionNode *action = 0, *actionParm;
    int state = 0, sepCount = 0, dimCount=0;

    // get keyword parameters
    while( (next_lexeme = statement->getNextLexeme()) ) {

        if(state == 0) {
            if(next_lexeme->type == Lexeme::type_identifier) {
                action = new ActionNode();
                action->name = "ERASE";
                actionParm = new ActionNode();
                actionParm->name = "_PARM_";
                actionParm->lexemes.push_back(next_lexeme);
                action->actions.push_back(actionParm);
                pushActionRoot(action);
                dimCount = 0;
                state ++;
                sepCount = 0;
            } else
                return false;
        } else if(state == 1) {
            if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == "(") {
                sepCount ++;
                state ++;
            } else if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == ",") {
                popActionRoot();
                action = 0;
                state = 0;
            } else
                return false;

        } else if(state == 2) {
            if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == "(") {
                sepCount ++;
            } else if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == ")") {
                sepCount --;
            }

            if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == ")" && sepCount == 0) {

                parm.setLexemeBOF();
                if(!eval_expression(&parm)) {
                    return false;
                }
                parm.clearLexemes();

                dimCount ++;

                if(dimCount == 1)
                    actionRoot->name = "REDIM1";
                else if(dimCount == 2)
                    actionRoot->name = "REDIM2";
                else
                    return false;

                popActionRoot();

                state ++;

            } else if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == ",") {

                dimCount ++;

                parm.setLexemeBOF();
                if(!eval_expression(&parm)) {
                    return false;
                }
                parm.clearLexemes();

            } else {
                parm.addLexeme(next_lexeme);
            }

        } else {
            if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == ",") {
                popActionRoot();
                action = 0;
                state = 0;
            } else
                return false;
        }

    }

    if(action)
        popActionRoot();

    return true;
}

bool Parser::eval_cmd_erase(LexerLine *statement) {
    Lexeme *next_lexeme;
    ActionNode *action = 0, *actionParm;
    int state = 0, sepCount = 0;

    // get keyword parameters
    while( (next_lexeme = statement->getNextLexeme()) ) {

        if(state == 0) {
            if(next_lexeme->type == Lexeme::type_identifier) {
                action = new ActionNode();
                action->name = "ERASE";
                actionParm = new ActionNode();
                actionParm->name = "_PARM_";
                actionParm->lexemes.push_back(next_lexeme);
                action->actions.push_back(actionParm);
                pushActionRoot(action);
                state ++;
                sepCount = 0;
            } else
                return false;
        } else if(state == 1) {
            if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == "(") {
                sepCount ++;
                state ++;
            } else if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == ",") {
                popActionRoot();
                action = 0;
                state = 0;
            } else
                return false;

        } else if(state == 2) {
            if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == "(") {
                sepCount ++;
            } else if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == ")") {
                sepCount --;
            }

            if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == ")" && sepCount == 0) {

                popActionRoot();

                state ++;

            }

        } else {
            if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == ",") {
                popActionRoot();
                action = 0;
                state = 0;
            } else
                return false;
        }

    }

    if(action)
        popActionRoot();

    return true;
}

bool Parser::eval_cmd_data(LexerLine *statement) {
    Lexeme *next_lexeme;

    if( (next_lexeme = statement->getNextLexeme()) ) {

        if(next_lexeme->type == Lexeme::type_keyword) {
            actionRoot->isRemark = true;
            actionRoot->name += "_" + next_lexeme->value;
            if       (next_lexeme->value == "SET") {
                return eval_cmd_data_set(statement);
            } else if(next_lexeme->value == "TILE") {
                return eval_cmd_data_tile(statement);
            } else if(next_lexeme->value == "SPRITE") {
                return eval_cmd_data_sprite(statement);
            } else if(next_lexeme->value == "SCREEN") {
                return eval_cmd_data_screen(statement);
            } else if(next_lexeme->value == "DECODER") {
                return eval_cmd_data_decoder(statement);
            } else if(next_lexeme->value == "MUSIC") {
                return eval_cmd_data_music(statement);
            }
        } else {
            next_lexeme = statement->getPreviousLexeme();
            return eval_cmd_generic(statement, -1);   // n-parameters
        }

    }

    return false;
}

bool Parser::eval_cmd_data_set(LexerLine *statement) {
    return eval_cmd_generic(statement, 1);
}

bool Parser::eval_cmd_data_music(LexerLine *statement) {
    Lexeme *next_lexeme, *lexeme;
    ActionNode *action;
    unsigned char *s;
    char snum[30];
    int i, t;

    if( (next_lexeme = statement->getNextLexeme()) ) {


        if(next_lexeme->value == "REPEAT") {

            t = 255;
            sprintf(snum, "%i", t);
            action = new ActionNode();
            action->name = "_PARM_";
            lexeme = new Lexeme();
            lexeme->name = lexeme->value = snum;
            lexeme->type = Lexeme::type_literal;
            lexeme->subtype = Lexeme::subtype_numeric;
            action->lexemes.push_back(lexeme);
            actionRoot->actions.push_back(action);

            return true;

        } else if(next_lexeme->type == Lexeme::type_literal) {

            t = next_lexeme->value.size();
            s = (unsigned char *) next_lexeme->value.c_str();

            for(i = 0; i <= t; i++) {

                if(s[i] == '"')
                    continue;

                sprintf(snum, "%i", s[i]);

                action = new ActionNode();
                action->name = "_PARM_";
                lexeme = new Lexeme();
                lexeme->name = lexeme->value = snum;
                lexeme->type = Lexeme::type_literal;
                lexeme->subtype = Lexeme::subtype_numeric;
                action->lexemes.push_back(lexeme);
                actionRoot->actions.push_back(action);

            }

            return true;

        }

    }

    return false;
}

bool Parser::eval_cmd_data_tile(LexerLine *statement) {
    Lexeme *next_lexeme;

    if( (next_lexeme = statement->getNextLexeme()) ) {

        if(next_lexeme->type == Lexeme::type_keyword) {
            actionRoot->name += "_" + next_lexeme->value;
            if       (next_lexeme->value == "SET") {
                return eval_cmd_generic(statement, -1);   // n-parameters
            } else if(next_lexeme->value == "COLOR") {
                return eval_cmd_generic(statement, -1);   // n-parameters
            } else if(next_lexeme->value == "INCLUDE") {
                return eval_cmd_data_include(statement, "TILE");
            } else if(next_lexeme->value == "LOAD") {
                return eval_cmd_data_load(statement);
            } else if(next_lexeme->value == "BANK") {
                return eval_cmd_generic(statement, 1);   // 1 parameter
            } else if(next_lexeme->value == "FONT") {
                return eval_cmd_generic(statement, 1);
            } else if(next_lexeme->value == "WALL") {
                return eval_cmd_generic(statement, -1);   // n-parameters
            } else if(next_lexeme->value == "HOTSPOT") {
                return eval_cmd_generic(statement, -1);   // n-parameters
            } else if(next_lexeme->value == "LIMIT") {
                return eval_cmd_data_tile_limit(statement);   // n-parameters
            }
        }

    }

    return false;
}

bool Parser::eval_cmd_data_tile_limit(LexerLine *statement) {
    Lexeme *next_lexeme;

    if( (next_lexeme = statement->getNextLexeme()) ) {

        if(next_lexeme->value == "ON") {
            actionRoot->name = "DATA_TILE_LIMIT_ON";
            return true;
        } else if(next_lexeme->value == "OFF") {
            actionRoot->name = "DATA_TILE_LIMIT_OFF";
            return true;
        }

    }

    return false;
}


bool Parser::eval_cmd_data_include(LexerLine *statement, string name) {
    Lexeme *next_lexeme, *lexeme;
    ActionNode *action;
    unsigned char *s;
    char filename[255], snum[30];
    char buffer[16365];
    int size_read, i, t, k;
    //bool color = false, sprite = false;

    if( (next_lexeme = statement->getNextLexeme()) ) {

        //color = (next_lexeme->value == "COLOR");
        //sprite = (name == "SPRITE");

        if(next_lexeme->value == "SET" || next_lexeme->value == "COLOR" || (next_lexeme->value == "MULTI" && name == "SPRITE")) {

            actionRoot->name = "DATA_" + name + "_" + next_lexeme->value;

            if( (next_lexeme = statement->getNextLexeme()) ) {

                if(next_lexeme->type == Lexeme::type_literal) {

                    t = next_lexeme->value.size();
                    s = (unsigned char *) next_lexeme->value.c_str();
                    k = 0;

                    for(i = 0; i < t; i++) {
                        if(s[i] != '"') {
                            filename[k] = s[i];
                            k++;
                        }
                    }
                    filename[k] = 0;

                    size_read = readFromFile(filename, &buffer[0], sizeof(buffer));

                    if(size_read) {
                        s = (unsigned char *) &buffer[0];
                        for(i = 0; i < size_read; i++) {
                            //if(sprite && color && s[i] > 15)
                            //    sprintf(snum, "%i", s[i] >> 4);  // TODO: transferir para o interpretador assembly essa decisão
                            //else
                            sprintf(snum, "%i", s[i]);
                            action = new ActionNode();
                            action->name = "_PARM_";
                            lexeme = new Lexeme();
                            lexeme->name = lexeme->value = snum;
                            lexeme->type = Lexeme::type_literal;
                            lexeme->subtype = Lexeme::subtype_numeric;
                            action->lexemes.push_back(lexeme);
                            actionRoot->actions.push_back(action);
                        }
                        return true;
                    }

                }

            }

        }

    }

    return false;
}

bool Parser::eval_cmd_data_load(LexerLine *statement) {
    Lexeme *next_lexeme;

    if( (next_lexeme = statement->getNextLexeme()) ) {

        if(next_lexeme->value == "SET" || next_lexeme->value == "COLOR") {

            actionRoot->name += next_lexeme->value;

            return eval_cmd_generic(statement, 1);

        }

    }

    return false;
}

bool Parser::eval_cmd_data_sprite(LexerLine *statement) {
    Lexeme *next_lexeme;

    if( (next_lexeme = statement->getNextLexeme()) ) {

        if(next_lexeme->type == Lexeme::type_keyword) {
            actionRoot->name += "_" + next_lexeme->value;
            if       (next_lexeme->value == "SET") {
                return eval_cmd_generic(statement, -1);   // n-parameters
            } else if(next_lexeme->value == "COLOR") {
                return eval_cmd_generic(statement, -1);   // n-parameters
            } else if(next_lexeme->value == "MULTI") {
                return eval_cmd_generic(statement, -1);   // n-parameters
            } else if(next_lexeme->value == "INCLUDE") {
                return eval_cmd_data_include(statement, "SPRITE");
            } else if(next_lexeme->value == "LOAD") {
                return eval_cmd_data_load(statement);
            }
        }

    }

    return false;
}

bool Parser::eval_cmd_data_screen(LexerLine *statement) {
    Lexeme *next_lexeme;

    if( (next_lexeme = statement->getNextLexeme()) ) {

        if(next_lexeme->type == Lexeme::type_keyword) {
            actionRoot->name += "_" + next_lexeme->value;
            if       (next_lexeme->value == "SET") {
                return eval_cmd_generic(statement, -1);   // n-parameters
            } else if(next_lexeme->value == "INCLUDE") {
                return eval_cmd_data_include(statement, "SCREEN");
            } else if(next_lexeme->value == "LOAD") {
                return eval_cmd_data_load(statement);
            }
        }

    }

    return false;
}

bool Parser::eval_cmd_data_decoder(LexerLine *statement) {
    return eval_cmd_generic(statement, 1);
}

bool Parser::eval_cmd_load(LexerLine *statement) {
    Lexeme *next_lexeme;

    if( (next_lexeme = statement->getNextLexeme()) ) {

        if(next_lexeme->type == Lexeme::type_keyword) {
            actionRoot->name += "_" + next_lexeme->value;
            if       (next_lexeme->value == "SET") {
                actionRoot->isRemark = true;
                return eval_cmd_generic(statement, 1);
            }
        }

    }

    return false;
}

bool Parser::eval_cmd_restore(LexerLine *statement) {
    return true;   // no parameters
}

bool Parser::eval_cmd_read(LexerLine *statement) {
    Lexeme *next_lexeme;
    LexerLine parm;
    ActionNode *action;

    // get keyword parameters
    while( (next_lexeme = statement->getNextLexeme()) ) {

        if(next_lexeme->type == Lexeme::type_separator && (next_lexeme->value == "," || next_lexeme->value == ";")) {

            action = new ActionNode();
            action->name = "READ";
            pushActionRoot(action);

            parm.setLexemeBOF();
            if(!eval_expression(&parm)) {
                return false;
            }

            popActionRoot();

            parm.clearLexemes();

        } else {
            parm.addLexeme(next_lexeme);
        }

    }

    if(parm.getLexemeCount()) {
        action = new ActionNode();
        action->name = "READ";
        pushActionRoot(action);

        parm.setLexemeBOF();
        if(!eval_expression(&parm)) {
            return false;
        }

        popActionRoot();
    }

    return true;
}

bool Parser::eval_cmd_end(LexerLine *statement) {
    return true;
}

bool Parser::eval_cmd_beep(LexerLine *statement) {
    return true;
}

bool Parser::eval_cmd_goto(LexerLine *statement) {
    actionRoot->isDirectParm = true;
    return eval_cmd_generic(statement, 1);
}

bool Parser::eval_cmd_gosub(LexerLine *statement) {
    actionRoot->isDirectParm = true;
    return eval_cmd_generic(statement, 1);
}

bool Parser::eval_cmd_return(LexerLine *statement) {
    return true;
}

bool Parser::eval_cmd_print(LexerLine *statement) {
    Lexeme *next_lexeme, *last_lexeme = 0;
    Lexeme *lexeme;
    LexerLine parm;
    ActionNode *action;
    int i;

    // get keyword parameters
    i = 0;
    while( (next_lexeme = statement->getNextLexeme()) ) {

        if(next_lexeme->type == Lexeme::type_separator && (next_lexeme->value == "," || next_lexeme->value == ";")) {

            action = new ActionNode();
            action->name = "PRINT";
            pushActionRoot(action);

            parm.setLexemeBOF();
            if(!eval_expression(&parm)) {
                return false;
            }

            popActionRoot();

            if(next_lexeme->value == ",") {
                action = new ActionNode();
                action->name = "PRINT";
                pushActionRoot(action);

                lexeme = new Lexeme();
                lexeme->name = "PRINT.TAB";
                lexeme->value = lexeme->name;
                lexeme->type = Lexeme::type_identifier;
                lexeme->subtype = Lexeme::subtype_internal;

                action = new ActionNode();
                action->name = "_PARM_";
                action->parmNo = ++i;
                action->lexemes.push_back(lexeme);
                actionRoot->actions.push_back(action);

                popActionRoot();
            }

            parm.clearLexemes();

        } else {
            parm.addLexeme(next_lexeme);
        }

        last_lexeme = next_lexeme;
    }

    if(parm.getLexemeCount()) {
        action = new ActionNode();
        action->name = "PRINT";
        pushActionRoot(action);

        parm.setLexemeBOF();
        if(!eval_expression(&parm)) {
            return false;
        }

        popActionRoot();
    }

    if(last_lexeme)
        if(last_lexeme->value != ";" && last_lexeme->value != ",") {
            action = new ActionNode();
            action->name = "PRINT";
            pushActionRoot(action);

            lexeme = new Lexeme();
            lexeme->name = "PRINT.CRLF";
            lexeme->value = lexeme->name;
            lexeme->type = Lexeme::type_identifier;
            lexeme->subtype = Lexeme::subtype_internal;

            action = new ActionNode();
            action->name = "_PARM_";
            action->parmNo = ++i;
            action->lexemes.push_back(lexeme);
            actionRoot->actions.push_back(action);

            popActionRoot();
        }

    return true;
}

bool Parser::eval_cmd_input(LexerLine *statement) {
    Lexeme *next_lexeme;
    Lexeme *lexeme;
    LexerLine parm;
    ActionNode *action;
    int i;

    // get keyword parameters
    i = 0;
    while( (next_lexeme = statement->getNextLexeme()) ) {

        if(next_lexeme->type == Lexeme::type_separator && (next_lexeme->value == "," || next_lexeme->value == ";")) {

            action = new ActionNode();
            action->name = "PRINT";
            if(parm.getLexemeCount() == 1) {
                lexeme = parm.getFirstLexeme();
                if(lexeme->type == Lexeme::type_identifier) {
                    if(lexeme->subtype == Lexeme::subtype_function || lexeme->tag == 1)
                        action->name = "INPUT.FUNCTION";
                    else
                        action->name = "INPUT";
                }
            }
            pushActionRoot(action);

            parm.setLexemeBOF();
            if(!eval_expression(&parm)) {
                return false;
            }

            popActionRoot();

            if(next_lexeme->value == ",") {
                action = new ActionNode();
                action->name = "PRINT";
                pushActionRoot(action);

                lexeme = new Lexeme();
                lexeme->name = "PRINT.TAB";
                lexeme->value = lexeme->name;
                lexeme->type = Lexeme::type_identifier;
                lexeme->subtype = Lexeme::subtype_internal;

                action = new ActionNode();
                action->name = "_PARM_";
                action->parmNo = ++i;
                action->lexemes.push_back(lexeme);
                actionRoot->actions.push_back(action);

                popActionRoot();
            }

            parm.clearLexemes();

        } else {
            parm.addLexeme(next_lexeme);
        }

    }

    if(parm.getLexemeCount()) {
        action = new ActionNode();
        action->name = "PRINT";
        if(parm.getLexemeCount() == 1) {
            lexeme = parm.getFirstLexeme();
            if(lexeme->type == Lexeme::type_identifier) {
                if(lexeme->subtype == Lexeme::subtype_function || lexeme->tag == 1)
                    action->name = "INPUT.FUNCTION";
                else
                    action->name = "INPUT";
            }
        }
        pushActionRoot(action);

        parm.setLexemeBOF();
        if(!eval_expression(&parm)) {
            return false;
        }

        popActionRoot();
    }

    return true;
}

bool Parser::eval_cmd_line_input(LexerLine *statement) {
    Lexeme *next_lexeme;
    Lexeme *lexeme;
    LexerLine parm;
    ActionNode *action;
    int i;

    // get keyword parameters
    i = 0;
    while( (next_lexeme = statement->getNextLexeme()) ) {

        if(next_lexeme->type == Lexeme::type_separator && (next_lexeme->value == "," || next_lexeme->value == ";")) {

            actionRoot->name = "LINE_INPUT";

            action = new ActionNode();
            action->name = "PRINT";
            pushActionRoot(action);

            parm.setLexemeBOF();
            if(!eval_expression(&parm)) {
                return false;
            }

            if(next_lexeme->value == ",") {
                action = new ActionNode();
                action->name = "PRINT";
                pushActionRoot(action);

                lexeme = new Lexeme();
                lexeme->name = "PRINT.TAB";
                lexeme->value = lexeme->name;
                lexeme->type = Lexeme::type_identifier;
                lexeme->subtype = Lexeme::subtype_internal;

                action = new ActionNode();
                action->name = "_PARM_";
                action->parmNo = ++i;
                action->lexemes.push_back(lexeme);
                actionRoot->actions.push_back(action);

                popActionRoot();
            }

            popActionRoot();

            parm.clearLexemes();

        } else {
            parm.addLexeme(next_lexeme);
        }

    }

    if(parm.getLexemeCount()) {
        actionRoot->name = "LINE_INPUT";

        parm.setLexemeBOF();
        if(!eval_expression(&parm)) {
            return false;
        }
    }

    return true;
}

bool Parser::eval_cmd_locate(LexerLine *statement) {
    return eval_cmd_generic(statement, 2);
}

bool Parser::eval_cmd_draw(LexerLine *statement) {
    // actionRoot->isBasicAction = true;
    actionRoot->name = "DO_DRAW";
    return eval_cmd_generic(statement, 1);
}

bool Parser::eval_cmd_play(LexerLine *statement) {
    Lexeme *next_lexeme;
    LexerLine parm;
    ActionNode *action;
    int i = 0;

    //    actionRoot->isBasicAction = true;
    //    return eval_cmd_generic(statement, 1);

    // get keyword parameters
    while( (next_lexeme = statement->getNextLexeme()) ) {

        if(next_lexeme->type == Lexeme::type_separator && (next_lexeme->value == "," || next_lexeme->value == ";")) {

            if(parm.getLexemeCount() == 0) {
                i ++;
                continue;
            }

            action = new ActionNode();
            if(i == 0)
                action->name = "SET_PLAY_VOICE_1";
            else if(i == 1)
                action->name = "SET_PLAY_VOICE_2";
            else if(i == 2)
                action->name = "SET_PLAY_VOICE_3";
            else
                return false;
            pushActionRoot(action);

            parm.setLexemeBOF();
            if(!eval_expression(&parm)) {
                return false;
            }

            popActionRoot();

            parm.clearLexemes();

            i++;

        } else {
            parm.addLexeme(next_lexeme);
        }

    }

    if(parm.getLexemeCount()) {

        action = new ActionNode();
        if(i == 0)
            action->name = "SET_PLAY_VOICE_1";
        else if(i == 1)
            action->name = "SET_PLAY_VOICE_2";
        else if(i == 2)
            action->name = "SET_PLAY_VOICE_3";
        else
            return false;
        pushActionRoot(action);

        parm.setLexemeBOF();
        if(!eval_expression(&parm)) {
            return false;
        }

        popActionRoot();

        i++;
    }

    if(i) {
        action = new ActionNode();
        action->name = "DO_PLAY";
        pushActionRoot(action);
        popActionRoot();
    }

    return true;
}

bool Parser::eval_cmd_music(LexerLine *statement) {
    Lexeme *next_lexeme;

    if( (next_lexeme = statement->getNextLexeme()) ) {

        if(next_lexeme->value == "PLAY") {
            actionRoot->name = "MUSIC_PLAY";
            return true;
            //} else if(next_lexeme->value == "NEXT") {
            //    actionRoot->name = "MUSIC_NEXT";
            //    return true;
        } else if(next_lexeme->value == "STOP") {
            actionRoot->name = "MUSIC_STOP";
            return true;
        }

    }

    return false;
}


bool Parser::eval_cmd_sound(LexerLine *statement) {
    return eval_cmd_generic(statement, 2);
}

bool Parser::eval_cmd_screen(LexerLine *statement) {
    Lexeme *next_lexeme;
    LexerLine parm;
    ActionNode *action;
    int i = 0;

    // get keyword parameters
    while( (next_lexeme = statement->getNextLexeme()) ) {

        if(next_lexeme->type == Lexeme::type_separator && (next_lexeme->value == "," || next_lexeme->value == ";")) {

            if(i == 0 && parm.getLexemeCount() == 0) {
                i ++;
                continue;
            }

            action = new ActionNode();
            if(i == 0)
                action->name = "SCREEN";
            else if(i == 1)
                action->name = "SPRITEMODE";
            else
                return false;
            pushActionRoot(action);

            parm.setLexemeBOF();
            if(!eval_expression(&parm)) {
                return false;
            }

            popActionRoot();

            parm.clearLexemes();

            i++;

        } else {
            parm.addLexeme(next_lexeme);
        }

    }

    if(parm.getLexemeCount()) {

        action = new ActionNode();
        if(i == 0)
            action->name = "SCREEN";
        else if(i == 1)
            action->name = "SPRITEMODE";
        else
            return false;
        pushActionRoot(action);

        parm.setLexemeBOF();
        if(!eval_expression(&parm)) {
            return false;
        }

        popActionRoot();
    }

    return true;
}

bool Parser::eval_cmd_color(LexerLine *statement) {
    Lexeme *next_lexeme, *last_lexeme=0;
    LexerLine parm;
    ActionNode *action;
    int i;
    bool first=true;

    i = 0;

    while( (next_lexeme = statement->getNextLexeme()) ) {

        if(first) {
            if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == "=") {
                return eval_cmd_color_rgb(statement);
            } else if(next_lexeme->type == Lexeme::type_keyword && next_lexeme->value == "SPRITE") {
                return eval_cmd_color_sprite(statement);
            } else
                first = false;
        }

        if(next_lexeme->type == Lexeme::type_separator && ( next_lexeme->value == "," || next_lexeme->value == ";")) {

            if(last_lexeme == 0 && i == 0) {
                i++;
                last_lexeme = next_lexeme;
                continue;
            } else if(last_lexeme)
                if(last_lexeme->value == ",") {
                    i++;
                    last_lexeme = next_lexeme;
                    continue;
                }

            action = new ActionNode();
            if( i == 0)
                action->name = "COLOR_FOREGROUND";
            else if( i == 1 )
                action->name = "COLOR_BACKGROUND";
            else if( i == 2 )
                action->name = "COLOR_BORDER";
            else {
                error_message = "Invalid number of parameters";
                return false;
            }
            pushActionRoot(action);

            parm.setLexemeBOF();
            if(!eval_expression(&parm)) {
                return false;
            }
            parm.clearLexemes();

            popActionRoot();
            i++;
        } else {
            last_lexeme = next_lexeme;
            parm.addLexeme(next_lexeme);
        }

    }

    if(parm.getLexemeCount()) {

        action = new ActionNode();
        if( i == 0)
            action->name = "COLOR_FOREGROUND";
        else if( i == 1 )
            action->name = "COLOR_BACKGROUND";
        else if( i == 2 )
            action->name = "COLOR_BORDER";
        else {
            error_message = "Invalid number of parameters";
            return false;
        }
        pushActionRoot(action);

        parm.setLexemeBOF();
        if(!eval_expression(&parm)) {
            return false;
        }

        popActionRoot();

        i++;
    }

    return true;
}

bool Parser::eval_cmd_color_rgb(LexerLine *statement) {
    Lexeme *next_lexeme;
    LexerLine parm;
    //ActionNode *action;
    int state = 0, sepCount=0;

    while( (next_lexeme = statement->getNextLexeme()) ) {

        if(state == 0) {
            if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == "(") {

                //action = new ActionNode();
                //action->name = "COLOR_RGB";
                //pushActionRoot(action);
                actionRoot->name += "_RGB";

                state ++;
                sepCount++;
                continue;

            } else if(next_lexeme->type == Lexeme::type_keyword && next_lexeme->value == "NEW") {

                //action = new ActionNode();
                //action->name = "COLOR_NEW";
                //pushActionRoot(action);
                //popActionRoot();
                actionRoot->name += "_NEW";
                return true;

            } else if(next_lexeme->type == Lexeme::type_keyword && next_lexeme->value == "NEW") {

                //action = new ActionNode();
                //action->name = "COLOR_RESTORE";
                //pushActionRoot(action);
                //popActionRoot();
                actionRoot->name += "_RESTORE";
                return true;

            } else
                return false;

        } else if(state == 1) {
            if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == "(") {
                sepCount ++;
            } else if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == ")") {
                sepCount --;
            }

            if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == ")" && sepCount == 0) {

                parm.setLexemeBOF();
                if(!eval_expression(&parm)) {
                    return false;
                }
                parm.clearLexemes();

                popActionRoot();

                return true;

            } else if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == ",") {

                parm.setLexemeBOF();
                if(!eval_expression(&parm)) {
                    return false;
                }
                parm.clearLexemes();

            } else {
                parm.addLexeme(next_lexeme);
            }


        }

    }

    return false;
}

bool Parser::eval_cmd_color_sprite(LexerLine *statement) {
    Lexeme *next_lexeme;
    LexerLine parm;
    //ActionNode *action;
    int state = 0, sepCount = 0;

    while( (next_lexeme = statement->getNextLexeme()) ) {

        if(state == 0) {
            if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == "(") {

                //action = new ActionNode();
                //action->name = "COLOR_SPRITE";
                //pushActionRoot(action);
                actionRoot->name += "_SPRITE";

                state ++;
                sepCount ++;
                continue;
            } else
                return false;

        } else if(state == 1) {
            if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == "(") {
                sepCount ++;
            } else if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == ")") {
                sepCount --;
            }

            if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == ")" && sepCount == 0) {

                parm.setLexemeBOF();
                if(!eval_expression(&parm)) {
                    return false;
                }
                parm.clearLexemes();

                state ++;
                continue;
            } else {
                parm.addLexeme(next_lexeme);
            }


        } else if(state == 2) {
            if(next_lexeme->type == Lexeme::type_operator && next_lexeme->value == "=") {
                state ++;
                continue;
            } else
                return false;

        } else {
            parm.addLexeme(next_lexeme);

        }

    }

    if(state < 3)
        return false;

    if(parm.getLexemeCount()) {
        parm.setLexemeBOF();
        if(!eval_expression(&parm)) {
            return false;
        }
    } else
        return false;

    //popActionRoot();

    return true;
}

bool Parser::eval_cmd_def(LexerLine *statement, int vartype) {
    Lexeme *next_lexeme;
    int state = 0, c[2], i;

    if(vartype == 0) {
        if((next_lexeme = statement->getNextLexeme())) {
            if( next_lexeme->name == "INT" )
                return eval_cmd_def(statement, 2);
            else if( next_lexeme->name == "STR" )
                return eval_cmd_def(statement, 3);
            else if( next_lexeme->name == "SNG" )
                return eval_cmd_def(statement, 4);
            else if( next_lexeme->name == "DBL" )
                return eval_cmd_def(statement, 8);
            else
                return false;
        } else
            return false;
    }

    c[0] = -1;
    c[1] = -1;

    while( (next_lexeme = statement->getNextLexeme()) ) {

        if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == ",") {

            if(state) {
                if(c[0] >= 0 && c[1] >= 0)
                    for(i = c[0]; i <= c[1]; i++)
                        deftbl[i] = vartype;
            } else {
                if(c[0] >= 0)
                    deftbl[c[0]] = vartype;
            }

            state = 0;
            c[0] = -1;
            c[1] = -1;

        } else if(next_lexeme->type == Lexeme::type_operator && next_lexeme->value == "-") {
            state = 1;

        } else if(next_lexeme->type == Lexeme::type_identifier) {
            c[state] = next_lexeme->name.c_str()[0] - 'A';
            if(!(c[state] >= 0 && c[state]<26))
                c[state] = -1;

        }

    }

    if(state) {
        if(c[0] >= 0 && c[1] >= 0)
            for(i = c[0]; i <= c[1]; i++)
                deftbl[i] = vartype;
    } else {
        if(c[0] >= 0)
            deftbl[c[0]] = vartype;
    }


    return true;
}


bool Parser::eval_cmd_put(LexerLine *statement) {
    Lexeme *next_lexeme;

    if( (next_lexeme = statement->getNextLexeme()) ) {

        if(next_lexeme->type == Lexeme::type_keyword) {
            if(next_lexeme->value == "SPRITE") {
                return eval_cmd_put_sprite(statement);
            } else if(next_lexeme->value == "KANJI") {
                return eval_cmd_put_kanji(statement);
            } else if(next_lexeme->value == "TILE") {
                return eval_cmd_put_tile(statement);
            }
        }

    }

    return false;
}

bool Parser::eval_cmd_put_sprite(LexerLine *statement) {
    Lexeme *next_lexeme;
    LexerLine parm;
    ActionNode *action;
    int state = 0, sepCount = 0, parmCount=0;

    while( (next_lexeme = statement->getNextLexeme()) ) {

        if(state == 0) {
            if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == ",") {

                action = new ActionNode();
                action->name = "PUT_SPRITE";
                pushActionRoot(action);

                parm.setLexemeBOF();
                if(!eval_expression(&parm)) {
                    return false;
                }
                parm.clearLexemes();

                state ++;
                continue;

            } else {
                parm.addLexeme(next_lexeme);
            }

        } else if(state == 1) {
            if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == "(") {
                state ++;
                sepCount ++;
                continue;
            } else if(next_lexeme->type == Lexeme::type_keyword && next_lexeme->value == "STEP") {
                actionRoot->name += "_STEP";
                continue;
            } else
                return false;


        } else if(state == 2) {
            if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == ",") {
                parm.setLexemeBOF();
                if(!eval_expression(&parm)) {
                    return false;
                }
                parm.clearLexemes();
                state ++;
                continue;
            } else {
                parm.addLexeme(next_lexeme);
            }

        } else if(state == 3) {
            if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == "(") {
                sepCount ++;
            } else if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == ")") {
                sepCount --;
            }
            if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == ")" && sepCount == 0) {
                parm.setLexemeBOF();
                if(!eval_expression(&parm)) {
                    return false;
                }
                parm.clearLexemes();
                continue;
            } else if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == ",") {
                parm.clearLexemes();
                state ++;
                continue;
            } else {
                parm.addLexeme(next_lexeme);
            }

        } else if(state == 4) {
            if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == ",") {
                if(parm.getLexemeCount()) {
                    parm.setLexemeBOF();
                    if(!eval_expression(&parm)) {
                        return false;
                    }
                    parm.clearLexemes();
                    if(parmCount==0)
                        actionRoot->name += "_COLOR";
                    else if(parmCount==1)
                        actionRoot->name += "_PATNUM";
                    else
                        return false;
                }
                parmCount ++;
                continue;
            } else {
                parm.addLexeme(next_lexeme);
            }

        }

    }

    if(state < 3)
        return false;

    if(parm.getLexemeCount()) {
        parm.setLexemeBOF();
        if(!eval_expression(&parm)) {
            return false;
        }
        if(state == 4) {
            if(parmCount==0)
                actionRoot->name += "_COLOR";
            else if(parmCount==1)
                actionRoot->name += "_PATNUM";
        }
    }

    popActionRoot();

    return true;
}

bool Parser::eval_cmd_put_kanji(LexerLine *statement) {
    return false;
}

bool Parser::eval_cmd_put_tile(LexerLine *statement) {
    Lexeme *next_lexeme;
    LexerLine parm;
    ActionNode *action;
    int state = 0, sepCount = 0, parmCount=0;

    while( (next_lexeme = statement->getNextLexeme()) ) {

        if(state == 0) {
            if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == ",") {

                action = new ActionNode();
                action->name = "PUT_TILE";
                pushActionRoot(action);

                parm.setLexemeBOF();
                if(!eval_expression(&parm)) {
                    return false;
                }
                parm.clearLexemes();

                state ++;
                continue;

            } else {
                parm.addLexeme(next_lexeme);
            }

        } else if(state == 1) {
            if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == "(") {
                state ++;
                sepCount ++;
                continue;
            } else
                return false;


        } else if(state == 2) {
            if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == ",") {
                parm.setLexemeBOF();
                if(!eval_expression(&parm)) {
                    return false;
                }
                parm.clearLexemes();
                state ++;
                continue;
            } else {
                parm.addLexeme(next_lexeme);
            }

        } else if(state == 3) {
            if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == "(") {
                sepCount ++;
            } else if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == ")") {
                sepCount --;
            }
            if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == ")" && sepCount == 0) {
                parm.setLexemeBOF();
                if(!eval_expression(&parm)) {
                    return false;
                }
                parm.clearLexemes();
                continue;
            } else if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == ",") {
                parm.clearLexemes();
                state ++;
                continue;
            } else {
                parm.addLexeme(next_lexeme);
            }

        } else if(state == 4) {
            if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == ",") {
                if(parm.getLexemeCount()) {
                    parm.setLexemeBOF();
                    if(!eval_expression(&parm)) {
                        return false;
                    }
                    parm.clearLexemes();
                    if(parmCount==0)
                        actionRoot->name += "_COLOR";
                    else
                        return false;
                }
                parmCount ++;
                continue;
            } else {
                parm.addLexeme(next_lexeme);
            }

        }

    }

    if(state < 3)
        return false;

    if(parm.getLexemeCount()) {
        parm.setLexemeBOF();
        if(!eval_expression(&parm)) {
            return false;
        }
        if(state == 4) {
            if(parmCount==0)
                actionRoot->name += "_COLOR";
        }
    }

    popActionRoot();

    return true;
}

bool Parser::eval_cmd_sprite(LexerLine *statement) {
    Lexeme *next_lexeme;
    LexerLine parm;
    ActionNode *action;
    int state = 0, sepCount = 0;

    while( (next_lexeme = statement->getNextLexeme()) ) {

        if(state == 0) {
            if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == "(") {

                action = new ActionNode();
                action->name = "SPRITE";
                pushActionRoot(action);

                state ++;
                sepCount ++;
                continue;

            } else if(next_lexeme->type == Lexeme::type_keyword && next_lexeme->value == "ON") {
                action = new ActionNode();
                action->name = "SPRITE_ON";
                pushActionRoot(action);
                popActionRoot();
                return true;
            } else if(next_lexeme->type == Lexeme::type_keyword && next_lexeme->value == "OFF") {
                action = new ActionNode();
                action->name = "SPRITE_OFF";
                pushActionRoot(action);
                popActionRoot();
                return true;
            } else if(next_lexeme->type == Lexeme::type_keyword && next_lexeme->value == "STOP") {
                action = new ActionNode();
                action->name = "SPRITE_STOP";
                pushActionRoot(action);
                popActionRoot();
                return true;
            } else
                return false;

        } else if(state == 1) {
            if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == "(") {
                sepCount ++;
            } else if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == ")") {
                sepCount --;
            }
            if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == ")" && sepCount == 0) {

                parm.setLexemeBOF();
                if(!eval_expression(&parm)) {
                    return false;
                }
                parm.clearLexemes();

                state ++;
                continue;
            } else {
                parm.addLexeme(next_lexeme);
            }


        } else if(state == 2) {
            if(next_lexeme->type == Lexeme::type_operator && next_lexeme->value == "=") {
                state ++;
                continue;
            } else
                return false;

        } else {
            parm.addLexeme(next_lexeme);

        }

    }

    if(state < 3)
        return false;

    if(parm.getLexemeCount()) {
        parm.setLexemeBOF();
        if(!eval_expression(&parm)) {
            return false;
        }
    } else
        return false;

    popActionRoot();

    return true;
}

bool Parser::eval_cmd_key(LexerLine *statement) {
    Lexeme *next_lexeme;
    LexerLine parm;
    ActionNode *action;
    int state = 0, sepCount = 0;

    while( (next_lexeme = statement->getNextLexeme()) ) {

        if(state == 0) {
            if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == "(") {

                state ++;
                sepCount ++;
                continue;

            } else
                return false;

        } else if(state == 1) {
            if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == "(") {
                sepCount ++;
            } else if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == ")") {
                sepCount --;
            }
            if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == ")" && sepCount == 0) {
                if(!parm.getLexemeCount())
                    return false;
                state ++;
                continue;
            } else {
                parm.addLexeme(next_lexeme);
            }

        } else if(state == 2) {
            action = new ActionNode();
            if(next_lexeme->type == Lexeme::type_keyword && next_lexeme->value == "ON") {
                action->name = "KEY_ON";
            } else if(next_lexeme->type == Lexeme::type_keyword && next_lexeme->value == "OFF") {
                action->name = "KEY_OFF";
            } else if(next_lexeme->type == Lexeme::type_keyword && next_lexeme->value == "STOP") {
                action->name = "KEY_STOP";
            } else
                return false;

            pushActionRoot(action);

            parm.setLexemeBOF();
            if(!eval_expression(&parm)) {
                return false;
            }
            parm.clearLexemes();

            popActionRoot();

            return true;

        }

    }

    return false;
}

bool Parser::eval_cmd_strig(LexerLine *statement) {
    Lexeme *next_lexeme;
    LexerLine parm;
    ActionNode *action;
    int state = 0, sepCount = 0;

    while( (next_lexeme = statement->getNextLexeme()) ) {

        if(state == 0) {
            if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == "(") {

                state ++;
                sepCount ++;
                continue;

            } else
                return false;

        } else if(state == 1) {
            if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == "(") {
                sepCount ++;
            } else if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == ")") {
                sepCount --;
            }
            if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == ")" && sepCount == 0) {
                if(!parm.getLexemeCount())
                    return false;
                state ++;
                continue;
            } else {
                parm.addLexeme(next_lexeme);
            }

        } else if(state == 2) {
            action = new ActionNode();
            if(next_lexeme->type == Lexeme::type_keyword && next_lexeme->value == "ON") {
                action->name = "STRIG_ON";
            } else if(next_lexeme->type == Lexeme::type_keyword && next_lexeme->value == "OFF") {
                action->name = "STRIG_OFF";
            } else if(next_lexeme->type == Lexeme::type_keyword && next_lexeme->value == "STOP") {
                action->name = "STRIG_STOP";
            } else
                return false;

            pushActionRoot(action);

            parm.setLexemeBOF();
            if(!eval_expression(&parm)) {
                return false;
            }
            parm.clearLexemes();

            popActionRoot();

            return true;

        }

    }

    return false;
}

bool Parser::eval_cmd_base(LexerLine *statement) {
    Lexeme *next_lexeme;
    LexerLine parm;
    ActionNode *action;
    int state = 0, sepCount = 0;

    while( (next_lexeme = statement->getNextLexeme()) ) {

        if(state == 0) {
            if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == "(") {

                action = new ActionNode();
                action->name = "SET_BASE";
                pushActionRoot(action);

                state ++;
                sepCount ++;
                continue;

            } else
                return false;

        } else if(state == 1) {
            if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == "(") {
                sepCount ++;
            } else if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == ")") {
                sepCount --;
            }
            if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == ")" && sepCount == 0) {

                parm.setLexemeBOF();
                if(!eval_expression(&parm)) {
                    return false;
                }
                parm.clearLexemes();

                state ++;
                continue;
            } else {
                parm.addLexeme(next_lexeme);
            }


        } else if(state == 2) {
            if(next_lexeme->type == Lexeme::type_operator && next_lexeme->value == "=") {
                state ++;
                continue;
            } else
                return false;

        } else {
            parm.addLexeme(next_lexeme);

        }

    }

    if(state < 3)
        return false;

    if(parm.getLexemeCount()) {
        parm.setLexemeBOF();
        if(!eval_expression(&parm)) {
            return false;
        }
    } else
        return false;

    popActionRoot();

    return true;
}

bool Parser::eval_cmd_vdp(LexerLine *statement) {
    Lexeme *next_lexeme;
    LexerLine parm;
    ActionNode *action;
    int state = 0, sepCount = 0;

    while( (next_lexeme = statement->getNextLexeme()) ) {

        if(state == 0) {
            if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == "(") {

                action = new ActionNode();
                action->name = "SET_VDP";
                pushActionRoot(action);

                state ++;
                sepCount ++;
                continue;

            } else
                return false;

        } else if(state == 1) {
            if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == "(") {
                sepCount ++;
            } else if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == ")") {
                sepCount --;
            }
            if(next_lexeme->type == Lexeme::type_separator && next_lexeme->value == ")" && sepCount == 0) {

                parm.setLexemeBOF();
                if(!eval_expression(&parm)) {
                    return false;
                }
                parm.clearLexemes();

                state ++;
                continue;
            } else {
                parm.addLexeme(next_lexeme);
            }


        } else if(state == 2) {
            if(next_lexeme->type == Lexeme::type_operator && next_lexeme->value == "=") {
                state ++;
                continue;
            } else
                return false;

        } else {
            parm.addLexeme(next_lexeme);

        }

    }

    if(state < 3)
        return false;

    if(parm.getLexemeCount()) {
        parm.setLexemeBOF();
        if(!eval_expression(&parm)) {
            return false;
        }
    } else
        return false;

    popActionRoot();

    return true;
}

bool Parser::eval_cmd_time(LexerLine *statement) {
    Lexeme *next_lexeme;
    LexerLine parm;
    ActionNode *action;
    int state = 0;

    while( (next_lexeme = statement->getNextLexeme()) ) {

        if(state == 0) {
            if(next_lexeme->type == Lexeme::type_operator && next_lexeme->value == "=") {

                action = new ActionNode();
                action->name = "SET_TIME";
                pushActionRoot(action);

                state ++;
                continue;

            } else
                return false;

        } else {
            parm.addLexeme(next_lexeme);

        }

    }

    if(parm.getLexemeCount()) {
        parm.setLexemeBOF();
        if(!eval_expression(&parm)) {
            return false;
        }
    } else
        return false;

    popActionRoot();

    return true;
}

bool Parser::eval_cmd_poke(LexerLine *statement) {
    return eval_cmd_generic(statement, 2);
}

bool Parser::eval_cmd_vpoke(LexerLine *statement) {
    return eval_cmd_generic(statement, 2);
}

bool Parser::eval_cmd_if(LexerLine *statement) {
    Lexeme *next_lexeme;
    LexerLine parm;
    ActionNode *action;
    int state = 0;
    bool imediate=false;

    if_count ++;

    actionRoot->name = "BOOLEAN.IF";

    action = new ActionNode();
    action->name = "_COND_";

    pushActionRoot(action);

    parm.clearLexemes();

    while( (next_lexeme = statement->getNextLexeme()) ) {

        if( next_lexeme->type == Lexeme::type_literal && next_lexeme->subtype == Lexeme::subtype_numeric
                && (actionRoot->name == "_THEN_" || actionRoot->name == "_ELSE_") && imediate ) {
            action = new ActionNode();
            action->name = "GOTO";
            action->isDirectParm = true;
            action->lexemes.push_back( next_lexeme );
            pushActionRoot(action);
            popActionRoot();
            continue;
        }

        imediate = false;

        if( next_lexeme->name == "THEN" && state == 0 ) {
            imediate = true;

            parm.setLexemeBOF();
            if( !eval_expression(&parm) ) {
                error_message = "IF command without a valid condition";
                eval_expr_error = true;
                return false;
            }

            parm.clearLexemes();
            popActionRoot();

            action = new ActionNode();
            action->name = "_THEN_";
            pushActionRoot(action);

            state ++;

            continue;

        } else if( next_lexeme->name == "ELSE" && state == 1 ) {

            imediate = true;

            if(actionRoot->name != "_THEN_") {
                error_message = "ELSE without a THEN";
                eval_expr_error = true;
                return false;
            }

            if(!eval_phrase(&parm))
                return false;

            parm.clearLexemes();
            popActionRoot();

            action = new ActionNode();
            action->name = "_ELSE_";
            pushActionRoot(action);

            state ++;

            continue;

        }

        parm.addLexeme(next_lexeme);
    }

    if(parm.getLexemeCount()) {

        if( actionRoot->name == "_COND_" ) {
            error_message = "IF command without a THEN/ELSE complement.";
            eval_expr_error = true;
            return false;
        }

        if(!eval_phrase(&parm))
            return false;

        parm.clearLexemes();

    }

    //popActionRoot();

    return true;
}

bool Parser::eval_cmd_then(LexerLine *statement) {
    return false;
}

bool Parser::eval_cmd_else(LexerLine *statement) {
    Lexeme *next_lexeme;
    LexerLine parm;
    ActionNode *action;
    bool imediate=false;

    imediate = true;

    parm.clearLexemes();
    popActionRoot();

    action = new ActionNode();
    action->name = "_ELSE_";
    pushActionRoot(action);

    while( (next_lexeme = statement->getNextLexeme()) ) {

        if( next_lexeme->type == Lexeme::type_literal && next_lexeme->subtype == Lexeme::subtype_numeric
                && (actionRoot->name == "_THEN_" || actionRoot->name == "_ELSE_") && imediate ) {
            action = new ActionNode();
            action->name = "GOTO";
            action->isDirectParm = true;
            action->lexemes.push_back( next_lexeme );
            pushActionRoot(action);
            popActionRoot();
            continue;
        }

        imediate = false;

        parm.addLexeme(next_lexeme);
    }

    if(parm.getLexemeCount()) {

        if(!eval_phrase(&parm))
            return false;

        parm.clearLexemes();

    }

    return true;
}

bool Parser::eval_cmd_for(LexerLine *statement) {
    Lexeme *next_lexeme, *identifier, *lexeme;
    LexerLine parm;
    ActionNode *action, *actionParm, *actionLet, *actionAdd, *actionStep, *actionCond;
    int state = 0;  // 0 = init, 1 = to, 2 = step

    // start of action INIT

    action = new ActionNode();
    action->name = "_INIT_";

    pushActionRoot(action);

    parm.clearLexemes();

    while( (next_lexeme = statement->getNextLexeme()) ) {

        if( next_lexeme->name == "TO" && state == 0 ) {

            // body of action INIT

            identifier = parm.getFirstLexeme();

            parm.setLexemeBOF();
            if( !eval_assignment(&parm) ) {
                error_message = "FOR command without a valid assigning expression";
                eval_expr_error = true;
                return false;
            }

            parm.clearLexemes();
            popActionRoot();

            // initialize the action STEP (it must occur before action TO in syntax tree)

            actionStep = new ActionNode();
            actionStep->name = "_STEP_";

            actionParm = new ActionNode();
            actionParm->name = "_PARM_";
            actionParm->parmNo = 0;
            actionParm->lexemes.push_back(identifier);

            actionAdd = new ActionNode();
            actionAdd->name = "MATH.ADD";
            actionAdd->parmNo = 0;
            actionAdd->actions.push_back(actionParm);   // (var + expr)

            actionLet = new ActionNode();
            actionLet->name = "LET";
            actionLet->actions.push_back(actionParm);   // let var = (var + expr )
            actionLet->actions.push_back(actionAdd);

            actionStep->actions.push_back(actionLet);

            actionRoot->actions.push_back(actionStep);

            // start of action TO

            action = new ActionNode();
            action->name = "_TO_";
            pushActionRoot(action);

            state ++;

            continue;

        } else if( next_lexeme->name == "STEP" && state == 1 ) {

            if(actionRoot->name != "_TO_") {
                error_message = "STEP without a FOR...TO";
                eval_expr_error = true;
                return false;
            }

            // body of action TO

            actionCond = new ActionNode();
            actionCond->name = "BOOLEAN.LE";
            pushActionRoot(actionCond);

            parm.setLexemeBOF();
            if( !eval_expression(&parm) ) {
                error_message = "FOR...TO command without a valid expression";
                eval_expr_error = true;
                return false;
            }

            popActionRoot();

            actionParm = new ActionNode();
            actionParm->name = "_PARM_";
            actionParm->parmNo = 0;
            actionParm->lexemes.push_back(identifier);   // variable <= (expression)?

            actionCond->actions.push_back(actionParm);

            parm.clearLexemes();
            popActionRoot();

            // start of action STEP

            state ++;

            continue;

        }

        parm.addLexeme(next_lexeme);
    }

    if(parm.getLexemeCount()) {

        if( state == 1 ) {   // state TO

            // body of action TO

            actionCond = new ActionNode();
            actionCond->name = "BOOLEAN.LE";
            pushActionRoot(actionCond);

            parm.setLexemeBOF();
            if( !eval_expression(&parm) ) {
                error_message = "FOR...TO command without a valid expression";
                eval_expr_error = true;
                return false;
            }

            popActionRoot();

            actionParm = new ActionNode();
            actionParm->name = "_PARM_";
            actionParm->parmNo = 0;
            actionParm->lexemes.push_back(identifier);   // variable <= (expression)?
            actionCond->actions.push_back(actionParm);

            popActionRoot();

            // body of action STEP

            lexeme = new Lexeme();
            lexeme->name = lexeme->value = "1";
            lexeme->type = Lexeme::type_literal;
            lexeme->subtype = Lexeme::subtype_numeric;

            actionParm = new ActionNode();
            actionParm->name = "_PARM_";
            actionParm->parmNo = 0;
            actionParm->lexemes.push_back(lexeme);
            if(actionAdd)
                actionAdd->actions.push_back(actionParm);

        } else if( state == 2 ) {   // state STEP

            // body of action STEP

            if(actionAdd) {
                action = actionRoot;        // save current root
                actionRoot = actionAdd;     // put actionAdd temporarily as root

                parm.setLexemeBOF();
                if( !eval_expression(&parm) ) {
                    error_message = "FOR...TO...STEP command without a valid expression";
                    eval_expr_error = true;
                    return false;
                }

                actionRoot = action;        // restore original root
            } else {
                error_message = "FOR...TO...STEP command without a valid expression";
                eval_expr_error = true;
                return false;
            }

        } else {
            error_message = "FOR command without a valid complement.";
            eval_expr_error = true;
            return false;
        }

    } else {
        error_message = "FOR command without a valid complement.";
        eval_expr_error = true;
        return false;
    }

    popActionRoot();

    return true;
}

bool Parser::eval_cmd_next(LexerLine *statement) {
    Lexeme *next_lexeme;
    ActionNode *actionParm;

    // test more next parameters

    while( (next_lexeme = statement->getNextLexeme()) ) {

        if( next_lexeme->type == Lexeme::type_separator && next_lexeme->value == "," ) {
            popActionRoot();
            actionParm = new ActionNode();
            actionParm->name = "NEXT";
            actionParm->parmNo = 0;
            pushActionRoot(actionParm);
        }

    }

    return true;
}

bool Parser::eval_cmd_sleep(LexerLine *statement) {
    return eval_cmd_generic(statement, 1);
}

bool Parser::eval_cmd_randomize(LexerLine *statement) {
    return true;
}

bool Parser::eval_cmd_seed(LexerLine *statement) {
    return eval_cmd_generic(statement, 1);
}

bool Parser::eval_cmd_pset(LexerLine *statement) {
    Lexeme *next_lexeme;
    ActionNode *actionParm;
    LexerLine parm;
    int state = 0, sepCount = 0, parmCount = 0;

    parm.clearLexemes();

    actionParm = new ActionNode();
    actionParm->name = "PSET.XY";
    actionParm->parmNo = 0;
    pushActionRoot(actionParm);

    while( (next_lexeme = statement->getNextLexeme()) ) {

        if( state == 0 ) {
            if(next_lexeme->type == Lexeme::type_identifier && next_lexeme->value == "STEP") {
                actionParm->name = "PSET.XY.STEP";
                continue;
            } else if( next_lexeme->type == Lexeme::type_separator && next_lexeme->value == "(") {
                state ++;
                sepCount ++;
                continue;
            } else {
                error_message = "PSET without a valid complement.";
                eval_expr_error = true;
                return false;
            }
        } else if(state == 1) {
            if( next_lexeme->type == Lexeme::type_separator && next_lexeme->value == "(" && parmCount < 3 ) {
                sepCount ++;
                if(sepCount <= 1)
                    continue;
            } else if( next_lexeme->type == Lexeme::type_separator && next_lexeme->value == ")" && parmCount < 3 ) {
                sepCount --;
                if(sepCount < 1)
                    continue;
            } else if( next_lexeme->type == Lexeme::type_separator && next_lexeme->value == "," ) {
                parmCount ++;
                if(parmCount == 3) {
                    popActionRoot();
                    actionParm = new ActionNode();
                    actionParm->name = "PSET.COLOR";
                    actionParm->parmNo = 0;
                    pushActionRoot(actionParm);
                }
                parm.setLexemeBOF();
                if(!eval_expression(&parm)) {
                    return false;
                }
                parm.clearLexemes();
                continue;
            }
            parm.addLexeme(next_lexeme);
        }

    }

    if(parm.getLexemeCount()) {
        parmCount ++;
        if(parmCount == 3) {
            popActionRoot();
            actionParm = new ActionNode();
            actionParm->name = "PSET.COLOR";
            actionParm->parmNo = 0;
            pushActionRoot(actionParm);
        }
        parm.setLexemeBOF();
        if(!eval_expression(&parm)) {
            return false;
        }
        parm.clearLexemes();
    }

    popActionRoot();

    if(parmCount < 2) {
        error_message = "PSET without enough parameters.";
        eval_expr_error = true;
        return false;
    } else if(parmCount > 3) {
        error_message = "PSET with excess of parameters.";
        eval_expr_error = true;
        return false;
    }

    return true;
}

bool Parser::eval_cmd_line(LexerLine *statement) {
    Lexeme *next_lexeme;
    ActionNode *actionParm = 0;
    LexerLine parm;
    int state = 0, sepCount = 0, parmCount = 0;
    bool xyStart = false, colorStart = false, secondStep = false;

    parm.clearLexemes();

    while( (next_lexeme = statement->getNextLexeme()) ) {

        if( state == 0 ) {
            if(next_lexeme->type == Lexeme::type_keyword && next_lexeme->value == "INPUT") {
                return eval_cmd_line_input(statement);
            } else if(next_lexeme->type == Lexeme::type_identifier && next_lexeme->value == "STEP") {
                if(!xyStart) {
                    actionParm = new ActionNode();
                    actionParm->name = "PSET.XY.STEP";
                    actionParm->parmNo = 0;
                    pushActionRoot(actionParm);
                    xyStart = true;
                } else
                    actionParm->name = "PSET.XY.STEP";
                continue;
            } else if( next_lexeme->type == Lexeme::type_separator && next_lexeme->value == "(") {
                state ++;
                sepCount ++;
                if(!xyStart) {
                    actionParm = new ActionNode();
                    actionParm->name = "PSET.XY";
                    actionParm->parmNo = 0;
                    pushActionRoot(actionParm);
                    xyStart = true;
                }
                continue;
            } else if( next_lexeme->type == Lexeme::type_operator && next_lexeme->value == "-" && parmCount == 0 ) {
                state ++;
                xyStart = false;
                parmCount = 2;
                continue;
            } else {
                error_message = "LINE without a valid complement.";
                eval_expr_error = true;
                return false;
            }
        } else if(state == 1) {
            if( next_lexeme->type == Lexeme::type_identifier && next_lexeme->value == "STEP" ) {
                secondStep = true;
                continue;
            } else if( next_lexeme->type == Lexeme::type_separator && next_lexeme->value == "("  && parmCount < 5 ) {
                sepCount ++;
                if(sepCount <= 1)
                    continue;
            } else if( next_lexeme->type == Lexeme::type_separator && next_lexeme->value == ")" && parmCount < 5 ) {
                sepCount --;
                if(sepCount < 1)
                    continue;
            } else if( ( next_lexeme->type == Lexeme::type_separator && next_lexeme->value == "," ) ||
                       ( next_lexeme->type == Lexeme::type_operator && next_lexeme->value == "-" && sepCount == 0  && parmCount == 1 ) ) {
                parmCount ++;
                if(parmCount == 3) {
                    if(xyStart) {
                        popActionRoot();
                        xyStart = false;
                    }
                } else if(parmCount == 5) {
                    if(xyStart) {
                        popActionRoot();
                        xyStart = false;
                    }
                    if(parm.getLexemeCount()) {
                        actionParm = new ActionNode();
                        actionParm->name = "PSET.COLOR";
                        actionParm->parmNo = 0;
                        pushActionRoot(actionParm);
                        colorStart = true;
                    }
                } else if(parmCount == 6) {
                    if(colorStart) {
                        popActionRoot();
                        colorStart = false;
                    }
                    next_lexeme = parm.getFirstLexeme();
                    if( next_lexeme->value == "B" ) {
                        actionRoot->name = "BOX";
                    } else if(next_lexeme->value == "BF") {
                        actionRoot->name = "FBOX";
                    } else
                        return false;
                    if( secondStep )
                        actionRoot->name += "_STEP";
                    parm.clearLexemes();
                    continue;
                }
                if(parm.getLexemeCount()) {
                    parm.setLexemeBOF();
                    if(!eval_expression(&parm)) {
                        return false;
                    }
                    parm.clearLexemes();
                }
                continue;
            }
            parm.addLexeme(next_lexeme);
        }

    }

    if(parm.getLexemeCount()) {
        parmCount ++;
        if(parmCount == 3) {
            if(xyStart) {
                popActionRoot();
                xyStart = false;
            }
        } else if(parmCount == 5) {
            if(xyStart) {
                popActionRoot();
                xyStart = false;
            }
            if(parm.getLexemeCount()) {
                actionParm = new ActionNode();
                actionParm->name = "PSET.COLOR";
                actionParm->parmNo = 0;
                pushActionRoot(actionParm);
                colorStart = true;
            }
        } else if(parmCount == 6) {
            if(colorStart) {
                popActionRoot();
                colorStart = false;
            }
            next_lexeme = parm.getFirstLexeme();
            if( next_lexeme->value == "B" ) {
                actionRoot->name = "BOX";
            } else if(next_lexeme->value == "BF") {
                actionRoot->name = "FBOX";
            } else
                return false;
            if( secondStep )
                actionRoot->name += "_STEP";
            parm.clearLexemes();
        }

        if(parm.getLexemeCount()) {
            parm.setLexemeBOF();
            if(!eval_expression(&parm)) {
                return false;
            }
            parm.clearLexemes();
        }
    }

    if(xyStart || colorStart)
        popActionRoot();

    if(parmCount < 4) {
        error_message = "LINE without enough parameters.";
        eval_expr_error = true;
        return false;
    } else if(parmCount > 6) {
        error_message = "LINE with excess of parameters.";
        eval_expr_error = true;
        return false;
    }

    return true;
}

bool Parser::eval_cmd_circle(LexerLine *statement) {
    Lexeme *next_lexeme;
    ActionNode *actionParm;
    LexerLine parm;
    int state = 0, sepCount = 0, parmCount = 0;
    bool xyStart = false, colorStart = false;

    parm.clearLexemes();

    actionParm = new ActionNode();
    actionParm->name = "PSET.XY";
    actionParm->parmNo = 0;
    pushActionRoot(actionParm);
    xyStart = true;

    while( (next_lexeme = statement->getNextLexeme()) ) {

        if( state == 0 ) {
            if(next_lexeme->type == Lexeme::type_identifier && next_lexeme->value == "STEP") {
                actionParm->name = "PSET.XY.STEP";
                continue;
            } else if( next_lexeme->type == Lexeme::type_separator && next_lexeme->value == "(") {
                state ++;
                sepCount ++;
                continue;
            } else {
                error_message = "CIRCLE without a valid complement.";
                eval_expr_error = true;
                return false;
            }
        } else if(state == 1) {
            if( next_lexeme->type == Lexeme::type_separator && next_lexeme->value == "(" && parmCount < 3 ) {
                sepCount ++;
                if(sepCount <= 1)
                    continue;
            } else if( next_lexeme->type == Lexeme::type_separator && next_lexeme->value == ")"  && parmCount < 3 ) {
                sepCount --;
                if(sepCount < 1)
                    continue;
            } else if( next_lexeme->type == Lexeme::type_separator && next_lexeme->value == "," ) {
                parmCount ++;
                if(parmCount == 3) {
                    if(xyStart)	{
                        popActionRoot();
                        xyStart = false;
                    }
                }
                if(parmCount == 4) {
                    actionParm = new ActionNode();
                    actionParm->name = "PSET.COLOR";
                    actionParm->parmNo = 0;
                    pushActionRoot(actionParm);
                    colorStart = true;
                }
                parm.setLexemeBOF();
                if(!eval_expression(&parm)) {
                    return false;
                }
                parm.clearLexemes();
                continue;
            }
            parm.addLexeme(next_lexeme);
        }

    }

    if(parm.getLexemeCount()) {
        parmCount ++;
        if(parmCount == 3) {
            if(xyStart)	{
                popActionRoot();
                xyStart = false;
            }
        }
        if(parmCount == 4) {
            actionParm = new ActionNode();
            actionParm->name = "PSET.COLOR";
            actionParm->parmNo = 0;
            pushActionRoot(actionParm);
            colorStart = true;
        }
        parm.setLexemeBOF();
        if(!eval_expression(&parm)) {
            return false;
        }
        parm.clearLexemes();
    }

    if(xyStart || colorStart)
        popActionRoot();

    if(parmCount < 3) {
        error_message = "CIRCLE without enough parameters.";
        eval_expr_error = true;
        return false;
    } else if(parmCount > 4) {
        error_message = "CIRCLE with excess of parameters.";
        eval_expr_error = true;
        return false;
    }

    return true;
}

bool Parser::eval_cmd_paint(LexerLine *statement) {
    Lexeme *next_lexeme;
    ActionNode *actionParm;
    LexerLine parm;
    int state = 0, sepCount = 0, parmCount = 0;
    bool xyStart = false, colorStart = false;

    parm.clearLexemes();

    actionParm = new ActionNode();
    actionParm->name = "PSET.XY";
    actionParm->parmNo = 0;
    pushActionRoot(actionParm);
    xyStart = true;

    while( (next_lexeme = statement->getNextLexeme()) ) {

        if( state == 0 ) {
            if(next_lexeme->type == Lexeme::type_identifier && next_lexeme->value == "STEP") {
                actionParm->name = "PSET.XY.STEP";
                continue;
            } else if( next_lexeme->type == Lexeme::type_separator && next_lexeme->value == "(") {
                state ++;
                sepCount ++;
                continue;
            } else {
                error_message = "PAINT without a valid complement.";
                eval_expr_error = true;
                return false;
            }
        } else if(state == 1) {
            if( next_lexeme->type == Lexeme::type_separator && next_lexeme->value == "(" ) {
                sepCount ++;
                if(sepCount < 1)
                    continue;
            } else if( next_lexeme->type == Lexeme::type_separator && next_lexeme->value == ")" ) {
                sepCount --;
                if(sepCount < 1)
                    continue;
            } else if( next_lexeme->type == Lexeme::type_separator && next_lexeme->value == "," ) {
                parmCount ++;
                if(parmCount == 3) {
                    if(xyStart)	{
                        popActionRoot();
                        xyStart = false;
                    }
                    actionParm = new ActionNode();
                    actionParm->name = "PSET.COLOR";
                    actionParm->parmNo = 0;
                    pushActionRoot(actionParm);
                    colorStart = true;
                }
                if(parmCount == 4) {
                    if(xyStart || colorStart)	{
                        popActionRoot();
                        xyStart = false;
                        colorStart = false;
                    }
                    actionParm = new ActionNode();
                    actionParm->name = "PAINT.BORDER";
                    actionParm->parmNo = 0;
                    pushActionRoot(actionParm);
                    colorStart = true;
                }
                parm.setLexemeBOF();
                if(!eval_expression(&parm)) {
                    return false;
                }
                parm.clearLexemes();
                continue;
            }
            parm.addLexeme(next_lexeme);
        }

    }

    if(parm.getLexemeCount()) {
        parmCount ++;
        if(parmCount == 3) {
            if(xyStart)	{
                popActionRoot();
                xyStart = false;
            }
            actionParm = new ActionNode();
            actionParm->name = "PSET.COLOR";
            actionParm->parmNo = 0;
            pushActionRoot(actionParm);
            colorStart = true;
        }
        if(parmCount == 4) {
            if(xyStart || colorStart)	{
                popActionRoot();
                xyStart = false;
                colorStart = false;
            }
            actionParm = new ActionNode();
            actionParm->name = "PAINT.BORDER";
            actionParm->parmNo = 0;
            pushActionRoot(actionParm);
            colorStart = true;
        }
        parm.setLexemeBOF();
        if(!eval_expression(&parm)) {
            return false;
        }
        parm.clearLexemes();
    }

    if(xyStart || colorStart)
        popActionRoot();

    if(parmCount < 2) {
        error_message = "PAINT without enough parameters.";
        eval_expr_error = true;
        return false;
    } else if(parmCount > 4) {
        error_message = "PAINT with excess of parameters.";
        eval_expr_error = true;
        return false;
    }

    return true;
}

bool Parser::eval_cmd_on(LexerLine *statement) {
    Lexeme *next_lexeme;

    has_traps = true;

    if( (next_lexeme = statement->getNextLexeme()) ) {

        if(next_lexeme->type == Lexeme::type_keyword) {
            if       (next_lexeme->value == "ERROR") {
                return eval_cmd_on_error(statement);
            } else if(next_lexeme->value == "INTERVAL") {
                return eval_cmd_on_interval(statement);
            } else if(next_lexeme->value == "KEY") {
                return eval_cmd_on_key(statement);
            } else if(next_lexeme->value == "SPRITE") {
                return eval_cmd_on_sprite(statement);
            } else if(next_lexeme->value == "STOP") {
                return eval_cmd_on_stop(statement);
            } else if(next_lexeme->value == "STRIG") {
                return eval_cmd_on_strig(statement);
            } else
                return false;
        } else  {
            return eval_cmd_on_goto_gosub(statement);
        }

    }

    return false;
}

bool Parser::eval_cmd_on_goto_gosub(LexerLine *statement) {
    Lexeme *next_lexeme, *lexeme;
    ActionNode *action, *actionParm;
    LexerLine parm;
    int i = 0;
    string goType;

    while( (next_lexeme = statement->getNextLexeme()) ) {

        if(next_lexeme->type == Lexeme::type_keyword && (next_lexeme->value == "GOSUB" || next_lexeme->value == "GOTO")) {

            if(! parm.getLexemeCount())
                return false;

            goType = next_lexeme->value;

            action = new ActionNode();
            action->name = "ON_" + goType + "_START";
            pushActionRoot(action);

            parm.setLexemeBOF();
            if(!eval_expression(&parm)) {
                return false;
            }

            popActionRoot();

            i++;

            while( (next_lexeme = statement->getNextLexeme()) ) {

                if(next_lexeme->type == Lexeme::type_literal) {

                    action = new ActionNode();
                    action->name = "ON_" + goType + "_NEXT";
                    pushActionRoot(action);

                    lexeme = new Lexeme();
                    lexeme->name = lexeme->value = "TAG_" + next_lexeme->value;
                    lexeme->type = Lexeme::type_literal;
                    lexeme->subtype = Lexeme::subtype_numeric;

                    actionParm = new ActionNode();
                    actionParm->name = "_PARM_";
                    actionParm->parmNo = 0;
                    actionParm->lexemes.push_back(lexeme);
                    actionRoot->actions.push_back(actionParm);

                    popActionRoot();

                    i ++;

                } else if(next_lexeme->type == Lexeme::type_separator && next_lexeme->name == ",") {
                    continue;
                } else
                    return false;

            }

        } else
            parm.addLexeme(next_lexeme);

    }

    return (i > 1);
}

bool Parser::eval_cmd_on_error(LexerLine *statement) {
    Lexeme *next_lexeme, *lexeme;
    ActionNode *action, *actionParm;

    if( (next_lexeme = statement->getNextLexeme()) ) {

        if(next_lexeme->type == Lexeme::type_keyword && next_lexeme->value == "GOTO") {

            action = new ActionNode();
            action->name = "ON_ERROR";
            pushActionRoot(action);

            if( (next_lexeme = statement->getNextLexeme()) ) {
                if(next_lexeme->type == Lexeme::type_literal) {
                    lexeme = new Lexeme();
                    lexeme->name = lexeme->value = "TAG_" + next_lexeme->value;
                    lexeme->type = Lexeme::type_literal;
                    lexeme->subtype = Lexeme::subtype_numeric;

                    actionParm = new ActionNode();
                    actionParm->name = "_PARM_";
                    actionParm->parmNo = 0;
                    actionParm->lexemes.push_back(lexeme);
                    actionRoot->actions.push_back(actionParm);
                } else
                    return false;
            } else
                return false;

            popActionRoot();

            return true;

        } else
            return false;

    }

    return false;
}

bool Parser::eval_cmd_on_interval(LexerLine *statement) {
    Lexeme *next_lexeme, *lexeme;
    ActionNode *action, *actionParm;
    LexerLine parm;
    bool first = true;

    while( (next_lexeme = statement->getNextLexeme()) ) {

        if(first) {
            first = false;
            if(next_lexeme->type == Lexeme::type_operator && next_lexeme->value == "=" ) {
                continue;
            }
        }

        if(next_lexeme->type == Lexeme::type_keyword && next_lexeme->value == "GOSUB") {

            if(! parm.getLexemeCount())
                return false;

            action = new ActionNode();
            action->name = "ON_INTERVAL";
            pushActionRoot(action);

            parm.setLexemeBOF();
            if(!eval_expression(&parm)) {
                return false;
            }

            if( (next_lexeme = statement->getNextLexeme()) ) {
                if(next_lexeme->type == Lexeme::type_literal) {
                    lexeme = new Lexeme();
                    lexeme->name = lexeme->value = "TAG_" + next_lexeme->value;
                    lexeme->type = Lexeme::type_literal;
                    lexeme->subtype = Lexeme::subtype_numeric;

                    actionParm = new ActionNode();
                    actionParm->name = "_PARM_";
                    actionParm->parmNo = 0;
                    actionParm->lexemes.push_back(lexeme);
                    actionRoot->actions.push_back(actionParm);
                } else
                    return false;
            } else
                return false;

            popActionRoot();

            return true;

        } else
            parm.addLexeme(next_lexeme);

    }

    return false;
}

bool Parser::eval_cmd_on_key(LexerLine *statement) {
    Lexeme *next_lexeme, *lexeme;
    ActionNode *action, *actionParm;
    int i = 0;

    if( (next_lexeme = statement->getNextLexeme()) ) {

        if(next_lexeme->type == Lexeme::type_keyword && next_lexeme->value == "GOSUB") {

            while( (next_lexeme = statement->getNextLexeme()) ) {

                if(next_lexeme->type == Lexeme::type_literal) {

                    action = new ActionNode();
                    if(i)
                        action->name = "ON_KEY_NEXT";
                    else
                        action->name = "ON_KEY_START";
                    pushActionRoot(action);

                    lexeme = new Lexeme();
                    lexeme->name = lexeme->value = "TAG_" + next_lexeme->value;
                    lexeme->type = Lexeme::type_literal;
                    lexeme->subtype = Lexeme::subtype_numeric;

                    actionParm = new ActionNode();
                    actionParm->name = "_PARM_";
                    actionParm->parmNo = 0;
                    actionParm->lexemes.push_back(lexeme);
                    actionRoot->actions.push_back(actionParm);

                    popActionRoot();

                    i ++;

                } else if(next_lexeme->type == Lexeme::type_separator && next_lexeme->name == ",") {
                    continue;
                } else
                    return false;

            }

        }

    }

    return (i > 0 && i <= 10);
}

bool Parser::eval_cmd_on_sprite(LexerLine *statement) {
    Lexeme *next_lexeme, *lexeme;
    ActionNode *action, *actionParm;

    if( (next_lexeme = statement->getNextLexeme()) ) {

        if(next_lexeme->type == Lexeme::type_keyword && next_lexeme->value == "GOSUB") {

            action = new ActionNode();
            action->name = "ON_SPRITE";
            pushActionRoot(action);

            if( (next_lexeme = statement->getNextLexeme()) ) {
                if(next_lexeme->type == Lexeme::type_literal) {
                    lexeme = new Lexeme();
                    lexeme->name = lexeme->value = "TAG_" + next_lexeme->value;
                    lexeme->type = Lexeme::type_literal;
                    lexeme->subtype = Lexeme::subtype_numeric;

                    actionParm = new ActionNode();
                    actionParm->name = "_PARM_";
                    actionParm->parmNo = 0;
                    actionParm->lexemes.push_back(lexeme);
                    actionRoot->actions.push_back(actionParm);
                } else
                    return false;
            } else
                return false;

            popActionRoot();

            return true;

        } else
            return false;

    }

    return false;
}

bool Parser::eval_cmd_on_stop(LexerLine *statement) {
    Lexeme *next_lexeme, *lexeme;
    ActionNode *action, *actionParm;

    if( (next_lexeme = statement->getNextLexeme()) ) {

        if(next_lexeme->type == Lexeme::type_keyword && next_lexeme->value == "GOSUB") {

            action = new ActionNode();
            action->name = "ON_STOP";
            pushActionRoot(action);

            if( (next_lexeme = statement->getNextLexeme()) ) {
                if(next_lexeme->type == Lexeme::type_literal) {
                    lexeme = new Lexeme();
                    lexeme->name = lexeme->value = "TAG_" + next_lexeme->value;
                    lexeme->type = Lexeme::type_literal;
                    lexeme->subtype = Lexeme::subtype_numeric;

                    actionParm = new ActionNode();
                    actionParm->name = "_PARM_";
                    actionParm->parmNo = 0;
                    actionParm->lexemes.push_back(lexeme);
                    actionRoot->actions.push_back(actionParm);
                } else
                    return false;
            } else
                return false;

            popActionRoot();

            return true;

        } else
            return false;

    }

    return false;
}

bool Parser::eval_cmd_on_strig(LexerLine *statement) {
    Lexeme *next_lexeme, *lexeme;
    ActionNode *action, *actionParm;
    int i = 0;

    if( (next_lexeme = statement->getNextLexeme()) ) {

        if(next_lexeme->type == Lexeme::type_keyword && next_lexeme->value == "GOSUB") {

            while( (next_lexeme = statement->getNextLexeme()) ) {

                if(next_lexeme->type == Lexeme::type_literal) {

                    action = new ActionNode();
                    if(i)
                        action->name = "ON_STRIG_NEXT";
                    else
                        action->name = "ON_STRIG_START";
                    pushActionRoot(action);

                    lexeme = new Lexeme();
                    lexeme->name = lexeme->value = "TAG_" + next_lexeme->value;
                    lexeme->type = Lexeme::type_literal;
                    lexeme->subtype = Lexeme::subtype_numeric;

                    actionParm = new ActionNode();
                    actionParm->name = "_PARM_";
                    actionParm->parmNo = 0;
                    actionParm->lexemes.push_back(lexeme);
                    actionRoot->actions.push_back(actionParm);

                    popActionRoot();

                    i ++;

                } else if(next_lexeme->type == Lexeme::type_separator && next_lexeme->name == ",") {
                    continue;
                } else
                    return false;

            }

        }

    }

    return (i > 0 && i <= 5);
}

bool Parser::eval_cmd_interval(LexerLine *statement) {
    Lexeme *next_lexeme;
    ActionNode *action;

    if( (next_lexeme = statement->getNextLexeme()) ) {

        if(next_lexeme->type == Lexeme::type_keyword && next_lexeme->value == "ON") {
            action = new ActionNode();
            action->name = "INTERVAL_ON";
            pushActionRoot(action);
            popActionRoot();
            return true;
        } else if(next_lexeme->type == Lexeme::type_keyword && next_lexeme->value == "OFF") {
            action = new ActionNode();
            action->name = "INTERVAL_OFF";
            pushActionRoot(action);
            popActionRoot();
            return true;
        } else if(next_lexeme->type == Lexeme::type_keyword && next_lexeme->value == "STOP") {
            action = new ActionNode();
            action->name = "INTERVAL_STOP";
            pushActionRoot(action);
            popActionRoot();
            return true;
        }

    }

    return false;
}

bool Parser::eval_cmd_stop(LexerLine *statement) {
    Lexeme *next_lexeme;
    ActionNode *action;

    if( (next_lexeme = statement->getNextLexeme()) ) {

        if(next_lexeme->type == Lexeme::type_keyword && next_lexeme->value == "ON") {
            action = new ActionNode();
            action->name = "STOP_ON";
            pushActionRoot(action);
            popActionRoot();
            return true;
        } else if(next_lexeme->type == Lexeme::type_keyword && next_lexeme->value == "OFF") {
            action = new ActionNode();
            action->name = "STOP_OFF";
            pushActionRoot(action);
            popActionRoot();
            return true;
        } else if(next_lexeme->type == Lexeme::type_keyword && next_lexeme->value == "STOP") {
            action = new ActionNode();
            action->name = "STOP_STOP";
            pushActionRoot(action);
            popActionRoot();
            return true;
        } else
            return false;

    } else {
        action = new ActionNode();
        action->name = "STOP";
        pushActionRoot(action);
        popActionRoot();
    }

    return true;
}

//-----------------------------------------------------------------------------------------------------------------

void Parser::pushActionRoot(ActionNode *action) {
    if(actionRoot) {
        actionRoot->actions.push_back(action);
        actionStack.push(actionRoot);
    } else {
        while(!actionStack.empty())
            actionStack.pop();
        tag->actions.push_back(action);
    }
    actionRoot = action;
}

void Parser::popActionRoot() {
    if(actionStack.empty()) {
        actionRoot = 0;
    } else {
        actionRoot = actionStack.top();
        actionStack.pop();
    }
}

int Parser::readFromFile(char *filename, char *data, int maxlen) {
    FILE *file;
    int total_bytes = 0;

    if ((file = fopen(filename, "r"))) {
        total_bytes = fread(data, 1, maxlen, file);
        fclose(file);
    }

    return total_bytes;
}

void ActionNode::print() {
    Lexeme *lexeme;
    ActionNode *action;
    unsigned int i;

    indent += 2;

    if(actions.size()) {

        printf("%*s{\n", indent, "");

        indent += 2;
    }

    if(lexemes.size()) {

        printf("%*s(\n", indent, "");

        indent += 2;

        for( i = 0; i < lexemes.size(); i++) {
            lexeme = lexemes[i];
            if(lexeme) {
                lexeme->indent = indent;
                lexeme->print();
            }
        }

        indent -= 2;

        printf("%*s) ", indent, "");

    }

    if(actions.size()) {
        for(i = 0; i < actions.size(); i++) {
            action = actions[i];
            if(action) {
                action->indent = indent;
                action->print();
            }
        }

        indent -= 2;

        printf("%*s} ", indent, "");
    } else if(!lexemes.size())
        printf("%*s", indent, "");

    printf("Action %s\n", name.c_str());

    indent -= 2;

}

void TagNode::print() {
    ActionNode *action;
    unsigned int i;

    printf("Tag %s\n", name.c_str());

    for(i = 0; i < actions.size(); i++) {
        action = actions[i];
        if(action)
            action->print();
    }
}

void Parser::print() {
    TagNode *tag;
    for(unsigned int i = 0; i < tags.size(); i++) {
        tag = tags[i];
        if(tag)
            tag->print();
    }
}

void Parser::error() {
    if(error_line)
        error_line->print();
    if(error_message.size() > 0)
        printf("%s\n", error_message.c_str());
}

ActionNode::ActionNode() {
    isBasicAction = false;
    isDirectParm = false;
    isMultiParm = false;
    isRemark = false;
    lastLexeme = 0;
    parmNo = 0;
    indent = 0;
}

